import enum
import logging
import time

from devtools import debug
from pydantic.v1 import (
    BaseModel,
    Field,
)

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ESRF.queue_entry.ssx_base_queue_entry import (
    BaseUserCollectionParameters,
    SsxBaseQueueEntry,
    SsxBaseQueueTaskParameters,
)
from mxcubecore.model.common import (
    CommonCollectionParamters,
    LegacyParameters,
    PathParameters,
    StandardCollectionParameters,
)
from mxcubecore.model.queue_model_objects import DataCollection

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class InjectorUserCollectionParameters(BaseUserCollectionParameters):
    exp_time: float = Field(100e-6, gt=0, lt=1)
    num_images: int = Field(1000, gt=0, lt=10000000)
    take_pedestal: bool = Field(True)
    sub_sampling: float = Field(4, gt=0, lt=100)

    class Config:
        extra: "ignore"
        use_enum_values: True


class InjectorColletionTaskParameters(SsxBaseQueueTaskParameters):
    path_parameters: PathParameters
    common_parameters: CommonCollectionParamters
    collection_parameters: StandardCollectionParameters
    user_collection_parameters: InjectorUserCollectionParameters
    legacy_parameters: LegacyParameters


class SsxInjectorCollectionQueueModel(DataCollection):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SsxInjectorCollectionQueueEntry(SsxBaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    QMO = SsxInjectorCollectionQueueModel
    DATA_MODEL = InjectorColletionTaskParameters
    NAME = "SSX Injector Collection"
    REQUIRES = ["point", "line", "no_shape", "chip", "mesh"]

    def __init__(self, view, data_model: SsxInjectorCollectionQueueModel):
        super().__init__(view=view, data_model=data_model)

    def execute(self):
        super().execute()

        exp_time = self._data_model._task_data.user_collection_parameters.exp_time
        fname_prefix = self._data_model._task_data.path_parameters.prefix
        num_images = self._data_model._task_data.user_collection_parameters.num_images
        sub_sampling = (
            self._data_model._task_data.user_collection_parameters.sub_sampling
        )
        self._data_model._task_data.collection_parameters.num_images = num_images
        data_root_path, _ = self.get_data_path()

        self.take_pedestal(HWR.beamline.collect.get_property("max_freq", 925))

        HWR.beamline.detector.prepare_acquisition(
            num_images, exp_time, data_root_path, fname_prefix
        )
        HWR.beamline.detector.wait_ready()

        self.start_processing("INJECTOR")

        HWR.beamline.diffractometer.set_phase("DataCollection")
        HWR.beamline.diffractometer.wait_ready()

        if HWR.beamline.control.safshut_oh2.state.name != "OPEN":
            logging.getLogger("user_level_log").info(f"Opening OH2 safety shutter")
            HWR.beamline.control.safshut_oh2.open()

        logging.getLogger("user_level_log").info(f"Acquiring ...")
        HWR.beamline.detector.start_acquisition()
        HWR.beamline.diffractometer.start_still_ssx_scan(num_images, sub_sampling)

        logging.getLogger("user_level_log").info(
            f"Waiting for acqusition to finish ..."
        )
        time.sleep(num_images * exp_time)

        HWR.beamline.diffractometer.wait_ready()
        HWR.beamline.detector.wait_ready()
        logging.getLogger("user_level_log").info(f"Acquired {num_images} images")

    def pre_execute(self):
        super().pre_execute()

    def post_execute(self):
        super().post_execute()

    def stop(self):
        super().stop()



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

""" Photon fluc calculations
Example xml file:
<object class="ESRF.ID30BPhotonFlux">
  <username>Photon flux</username>
  <object role="controller" href="/bliss"/>
  <object role="aperture" href="/udiff_aperture"/>
  <counter_name>i0</counter_name>
</object>
"""
import logging

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractFlux import AbstractFlux


class ID30BPhotonFlux(AbstractFlux):
    """Photon flux calculation for ID30B"""

    def __init__(self, name):
        super(ID30BPhotonFlux, self).__init__(name)
        self._counter = None
        self._flux_calc = None
        self._aperture = None

    def init(self):
        """Initialisation"""
        super(ID30BPhotonFlux, self).init()
        controller = self.get_object_by_role("controller")
        # ultimately it should be HWR.beamline.diffractometer.aperture
        self._aperture = self.get_object_by_role("aperture")

        try:
            self._flux_calc = controller.CalculateFlux()
            self._flux_calc.init()
        except AttributeError:
            logging.getLogger("HWR").exception(
                "Could not get flux calculation from BLISS"
            )

        counter = self.get_property("counter_name")
        if counter:
            self._counter = getattr(controller, counter)
        else:
            self._counter = self.get_object_by_role("counter")

        HWR.beamline.safety_shutter.connect("stateChanged", self.update_value)

    def get_value(self):
        """Calculate the flux value as function of a reading
        """
        # counts = self._counter._counter_controller.read_all(self._counter)
        # if isinstance(counts, list):
        #    counts = float(counts[0])
        counts = float(self._counter.raw_read)
        if counts == -9999:
            counts = 0.0

        egy = HWR.beamline.energy.get_value() * 1000.0
        calib = self._flux_calc.calc_flux_factor(egy)[self._counter.name]

        try:
            label = self._aperture.get_value().name
            aperture_factor = self._aperture.get_factor(label)
        except AttributeError:
            aperture_factor = 1
        counts = abs(counts * calib * aperture_factor)
        self._nominal_value = counts
        return counts



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
AbstractBeam class - methods to define the size and shape of the beam.

emits:
- beamSizeChanged (self._beam_width, self._beam_height)
- beamInfoChanged (self._beam_info_dict.copy())
"""

__copyright__ = """ Copyright © by MXCuBE Collaboration """
__license__ = "LGPLv3+"


import abc
import sys
from enum import (
    Enum,
    unique,
)
from warnings import warn

from mxcubecore.BaseHardwareObjects import HardwareObject


@unique
class BeamShape(Enum):
    """Beam shape definitions"""

    UNKNOWN = "unknown"
    RECTANGULAR = "rectangular"
    ELLIPTICAL = "ellipse"


class AbstractBeam(HardwareObject):
    """AbstractBeam class"""

    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        super().__init__(name)

        self._aperture = None
        self._slits = None
        self._definer = None
        self._definer_type = None

        self._beam_size_dict = {
            "aperture": [sys.float_info.max, sys.float_info.max],
            "slits": [sys.float_info.max, sys.float_info.max],
            "definer": [sys.float_info.max, sys.float_info.max],
        }
        self._beam_width = None
        self._beam_height = None
        self._beam_shape = BeamShape.UNKNOWN
        self._beam_label = None
        self._beam_divergence = (None, None)
        self._beam_position_on_screen = [None, None]

        self._beam_info_dict = {
            "size_x": self._beam_width,
            "size_y": self._beam_height,
            "shape": self._beam_shape,
            "label": self._beam_label,
        }

    def init(self):
        """
        Initialise default values and objects
        """
        super().init()
        _divergence_vertical = self.get_property("beam_divergence_vertical")
        _divergence_horizontal = self.get_property("beam_divergence_horizontal")
        self._beam_divergence = (_divergence_horizontal, _divergence_vertical)
        self._beam_position_on_screen = [0, 0]
        self._definer_type = self.get_property("definer_type")

    @property
    def aperture(self):
        """
        Returns aperture hwobj
        """
        return self._aperture

    @property
    def slits(self):
        """
        Returns slits hwobj
        """
        return self._slits

    @property
    def definer(self):
        """
        Beam definer device, equipment like focusing optics, CRLs, and etc.
        """
        return self._definer

    def get_beam_divergence(self):
        """Get the beam divergence.
        Returns:
            (tuple): Beam divergence (horizontal, vertical) [μm]
        """
        return self._beam_divergence

    def get_available_size(self):
        """Get the available beam definers configuration.
        Returns:
            (dict): Dictionary {"type": (list), "values": (list)}, where
               "type": the definer type ("aperture", "slits","definer")
               "values": List of available beam size difinitions,
                         according to the "type".
        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def get_defined_beam_size(self):
        """Get the predefined beam labels and size.
        Returns:
            (dict): Dictionary wiith list of avaiable beam size labels
                    and the corresponding size (width,height) tuples.
                    {"label": [str, str, ...], "size": [(w,h), (w,h), ...]}
        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def get_beam_shape(self):
        """
        Returns:
            beam_shape: Enum BeamShape
        """
        self.evaluate_beam_info()
        return self._beam_shape

    def get_beam_size(self):
        """
        Returns:
            (tuple): two floats
        """
        self.evaluate_beam_info()
        return self._beam_width, self._beam_height

    def set_value(self, size=None):
        """Set the beam size.
        Args:
            size (list): Width, heigth [um] or
                  (str): Aperture or definer name.
        Raises:
            NotImplementedError
        """
        raise NotImplementedError

    def set_beam_size_shape(self, beam_width, beam_height, beam_shape):
        """
        Sets beam size and shape
        Args:
            beam_width (float): requested beam width in microns
            beam_height (float): requested beam height in microns
            beam_shape (BeamShape enum): requested beam shape
        """
        warn(
            "set_beam_size_shape is deprecated. Use set_value() instead",
            DeprecationWarning,
        )

        if beam_shape == BeamShape.RECTANGULAR:
            self._slits.set_horizontal_gap(beam_width)
            self._slits.set_vertical_gap(beam_height)
        elif beam_shape == BeamShape.ELLIPTICAL:
            self._aperture.set_diameter_size(beam_width)

    def get_beam_position_on_screen(self):
        """Get the beam position
        Returns:
            (tuple): Position (x, y) [pixel]
        """
        # (TODO) move this method to AbstractSampleView
        return self._beam_position_on_screen

    def set_beam_position_on_screen(self, beam_x_y):
        """Set the beam position
        Returns:
            beam_x_y (tuple): Position (x, y) [pixel]
        """
        raise NotImplementedError

    def get_beam_info_dict(self):
        """
        Returns:
            (dict): copy of beam_info_dict
        """
        return self._beam_info_dict.copy()

    def evaluate_beam_info(self):
        """
        Method called if aperture, slits or focusing has been changed.
        Evaluates which of the beam size defining devices determins the size.
        Returns:
            (dict): Beam info dictionary (dict), type of the definer (str).
                     {size_x: float, size_y: float,
                      shape: BeamShape enum, label: str},
        """
        _shape = BeamShape.UNKNOWN
        _size = min(self._beam_size_dict.values())
        key = [k for k, v in self._beam_size_dict.items() if v == _size]

        if len(key) == 1:
            _label = key[0]
        else:
            if self._definer_type in key:
                _label = self._definer_type
            else:
                _label = "UNKNOWN"

        if _label == "slits":
            _shape = BeamShape.RECTANGULAR
        else:
            _shape = BeamShape.ELLIPTICAL

        self._beam_width = _size[0]
        self._beam_height = _size[1]
        self._beam_shape = _shape
        self._beam_info_dict["size_x"] = _size[0]
        self._beam_info_dict["size_y"] = _size[1]
        self._beam_info_dict["shape"] = _shape
        self._beam_info_dict["label"] = _label

        return self._beam_info_dict

    def re_emit_values(self):
        """
        Reemits beamSizeChanged and beamInfoChanged signals
        """
        HardwareObject.re_emit_values(self)
        if self._beam_width != 9999 and self._beam_height != 9999:
            self.emit("beamSizeChanged", (self._beam_width, self._beam_height))
            self.emit("beamInfoChanged", (self._beam_info_dict))
            self.emit("beamPosChanged", (self._beam_position_on_screen,))



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Abstract Actuator class.
Defines the set/update value, get/set/update limits and validate_value
methods and the get_value and _set_value abstract methods.
Initialises the actuator_name, username, read_only and default_value properties.
Emits signals valueChanged and limitsChanged.
"""

import abc
import math
from ast import literal_eval

from mxcubecore.BaseHardwareObjects import HardwareObject

__copyright__ = """ Copyright © 2010-2022 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class AbstractActuator(HardwareObject):
    """Abstract actuator"""

    __metaclass__ = abc.ABCMeta

    unit = None

    def __init__(self, name):
        super().__init__(name)
        self._nominal_value = None
        self._nominal_limits = (None, None)
        self.actuator_name = None
        self.read_only = False
        self.default_value = None
        self.username = None

    def init(self):
        """Initialise actuator_name, username, read_only and default_value
        properties.
        """
        self.actuator_name = self.get_property("actuator_name")
        self.read_only = self.get_property("read_only") or False
        self.default_value = self.get_property("default_value")
        if self.default_value is not None:
            self.update_value(self.default_value)
        limits = self.get_property("default_limits")
        if limits:
            try:
                self._nominal_limits = tuple(literal_eval(limits))
            except TypeError:
                print("Invalid limits")
        self.username = self.get_property("username")

    @abc.abstractmethod
    def get_value(self):
        """Read the actuator position.
        Returns:
            value: Actuator position.
        """
        return None

    def get_limits(self):
        """Return actuator low and high limits.
        Returns:
            (tuple): two elements (low limit, high limit) tuple.
        """
        return self._nominal_limits

    def set_limits(self, limits):
        """Set actuator low and high limits. Emits signal limitsChanged.
        Args:
            limits (tuple): two elements (low limit, high limit) tuple.
        Raises:
            ValueError: Attempt to set limits for read-only Actuator.
        """
        if self.read_only:
            raise ValueError("Attempt to set limits for read-only Actuator")

        self._nominal_limits = limits
        self.emit("limitsChanged", (self._nominal_limits,))

    def validate_value(self, value):
        """Check if the value is within limits.
        Args:
            value(numerical): value
        Returns:
            (bool): True if within the limits
        """
        if value is None:
            return True
        if math.isnan(value) or math.isinf(value):
            return False
        if None in self._nominal_limits:
            return True
        return self._nominal_limits[0] <= value <= self._nominal_limits[1]

    @abc.abstractmethod
    def _set_value(self, value):
        """Implementation of specific set actuator logic.
        Args:
            value: target value
        """

    def set_value(self, value, timeout=0):
        """Set actuator to value.
        Args:
            value: target value
            timeout (float): optional - timeout [s],
                             If timeout == 0: return at once and do not wait
                                              (default);
                             if timeout is None: wait forever.
        Raises:
            ValueError: Invalid value or attemp to set read only actuator.
            RuntimeError: Timeout waiting for status ready  # From wait_ready
        """
        if self.read_only:
            raise ValueError("Attempt to set value for read-only Actuator")
        if self.validate_value(value):
            self._set_value(value)
            self.update_value()
            if timeout == 0:
                return
            self.wait_ready(timeout)
        else:
            raise ValueError(f"Invalid value {value}")

    def update_value(self, value=None):
        """Check if the value has changed. Emits signal valueChanged.
        Args:
            value: value
        """
        if value is None:
            value = self.get_value()

        if self._nominal_value != value:
            self._nominal_value = value
            self.emit("valueChanged", (value,))

    def update_limits(self, limits=None):
        """Check if the limits have changed. Emits signal limitsChanged.
        Args:
            limits (tuple): two elements tuple (low limit, high limit).
        """
        if not limits:
            limits = self.get_limits()

        if self._nominal_limits != limits:
            # All values are not NaN
            if not any(isinstance(lim, float) and math.isnan(lim) for lim in limits):
                self._nominal_limits = limits
                self.emit("limitsChanged", (limits,))

    def re_emit_values(self):
        """Update values for all internal attributes"""
        self.update_value(self.get_value())
        self.update_limits(self.get_limits())
        super(AbstractActuator, self).re_emit_values()

    def force_emit_signals(self):
        """Forces to emit all signals.

        Method is called from gui
        Do not call it within HWR
        """
        self.emit("valueChanged", (self.get_value(),))
        self.emit("limitsChanged", (self.get_limits(),))
        self.emit("stateChanged", (self.get_state(),))



import json
import logging
import math

from devtools import debug
from pydantic import Field
from typing_extensions import Literal

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ESRF.queue_entry.ssx_base_queue_entry import (
    BaseUserCollectionParameters,
    SsxBaseQueueEntry,
    SsxBaseQueueTaskParameters,
)
from mxcubecore.model.common import (
    CommonCollectionParamters,
    LegacyParameters,
    PathParameters,
    StandardCollectionParameters,
)
from mxcubecore.model.queue_model_objects import DataCollection

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3+"
__category__ = "General"


class SSXUserCollectionParameters(BaseUserCollectionParameters):
    horizontal_spacing: float = Field(8, gt=0, lt=1000, description="um")
    vertical_spacing: float = Field(8, gt=0, lt=1000, description="um")

    try:
        _chip_name_tuple = tuple(
            HWR.beamline.diffractometer.get_head_configuration().available.keys()
        )
        _current_chip = HWR.beamline.diffractometer.get_head_configuration().current
    except AttributeError:
        _chip_name_tuple = tuple("")
        _current_chip = ""

    chip_type: Literal[_chip_name_tuple] = Field(_current_chip)

    class Config:
        extra: "ignore"


class SsxFoilColletionTaskParameters(SsxBaseQueueTaskParameters):
    path_parameters: PathParameters
    common_parameters: CommonCollectionParamters
    collection_parameters: StandardCollectionParameters
    user_collection_parameters: SSXUserCollectionParameters
    legacy_parameters: LegacyParameters

    @staticmethod
    def ui_schema():
        schema = json.loads(SsxBaseQueueTaskParameters.ui_schema())
        schema.update(
            {
                "sub_sampling": {"ui:readonly": "true"},
            }
        )
        return json.dumps(schema)

    @staticmethod
    def update_dependent_fields(field_data):
        horizontal_spacing = field_data["horizontal_spacing"]
        vertical_spacing = field_data["vertical_spacing"]
        sub_sampling = field_data["sub_sampling"]
        chip_type = field_data["chip_type"]

        chip_name = chip_type
        chip_data = HWR.beamline.diffractometer.get_head_configuration().available[
            chip_name
        ]

        chip_width = (
            chip_data.calibration_data.top_right[0]
            - chip_data.calibration_data.top_left[0]
        )
        chip_height = (
            chip_data.calibration_data.bottom_left[1]
            - chip_data.calibration_data.top_left[1]
        )

        nb_samples_per_line = math.floor(
            chip_width / ((horizontal_spacing / 1000) * sub_sampling)
        )
        nb_lines = math.floor(chip_height / (vertical_spacing / 1000))

        num_images = math.floor((nb_samples_per_line * nb_lines) / 2) * 2

        new_data = {"num_images": num_images}
        return new_data


class SsxFoilCollectionQueueModel(DataCollection):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class SsxFoilCollectionQueueEntry(SsxBaseQueueEntry):
    """
    Defines the behaviour of a data collection.
    """

    QMO = SsxFoilCollectionQueueModel
    DATA_MODEL = SsxFoilColletionTaskParameters
    NAME = "SSX Foil Collection"
    REQUIRES = ["point", "line", "no_shape", "chip", "mesh"]

    # New style queue entry does not take view argument,
    # adding kwargs for compatability, but they are unsued
    def __init__(self, view, data_model: SsxFoilCollectionQueueModel):
        super().__init__(view=view, data_model=data_model)

    def execute(self):
        super().execute()
        debug(self._data_model._task_data)
        params = self._data_model._task_data.user_collection_parameters
        packet_fifo_depth = 20000

        chip_name = params.chip_type
        chip_data = HWR.beamline.diffractometer.get_head_configuration().available[
            chip_name
        ]

        chip_width = (
            chip_data.calibration_data.top_right[0]
            - chip_data.calibration_data.top_left[0]
        )
        chip_height = (
            chip_data.calibration_data.bottom_left[1]
            - chip_data.calibration_data.top_left[1]
        )

        nb_samples_per_line = math.floor(
            chip_width / ((params.horizontal_spacing / 1000) * params.sub_sampling)
        )
        nb_lines = math.floor(chip_height / (params.vertical_spacing / 1000))

        exp_time = self._data_model._task_data.user_collection_parameters.exp_time
        num_images = math.floor((nb_samples_per_line * nb_lines) / 2) * 2

        self._data_model._task_data.collection_parameters.num_images = num_images
        fname_prefix = self._data_model._task_data.path_parameters.prefix
        data_root_path, _ = self.get_data_path()

        self.take_pedestal(HWR.beamline.collect.get_property("max_freq", 925))

        logging.getLogger("user_level_log").info("Preparing detector")
        HWR.beamline.detector.prepare_acquisition(
            num_images, exp_time, data_root_path, fname_prefix
        )

        HWR.beamline.detector.wait_ready()

        fname_prefix = self._data_model._task_data.path_parameters.prefix
        fname_prefix += f"_foil_"

        region = (
            chip_data.calibration_data.top_left[0],
            chip_data.calibration_data.top_left[1],
            chip_data.calibration_data.top_left[2],
            chip_data.calibration_data.top_right[0],
            chip_data.calibration_data.top_right[1],
            chip_data.calibration_data.top_right[2],
            chip_data.calibration_data.bottom_left[0],
            chip_data.calibration_data.bottom_left[1],
            chip_data.calibration_data.bottom_left[2],
        )

        self.start_processing("FOIL")

        logging.getLogger("user_level_log").info(f"Defining region {region}")

        HWR.beamline.diffractometer.define_ssx_scan_region(
            *region, nb_samples_per_line, nb_lines
        )

        HWR.beamline.diffractometer.wait_ready()
        HWR.beamline.diffractometer.set_phase("DataCollection", wait=True, timeout=120)

        if HWR.beamline.control.safshut_oh2.state.name != "OPEN":
            logging.getLogger("user_level_log").info(f"Opening OH2 safety shutter")
            HWR.beamline.control.safshut_oh2.open()

        HWR.beamline.detector.start_acquisition()
        logging.getLogger("user_level_log").info(
            "Detector ready, waiting for trigger ..."
        )

        logging.getLogger("user_level_log").info(f"Acquiring region {region}")
        logging.getLogger("user_level_log").info(
            f"Sub sampling is {params.sub_sampling}"
        )
        logging.getLogger("user_level_log").info(
            f"Acquiring {num_images} images ({nb_lines} lines x {nb_samples_per_line} samples per line)"
        )
        logging.getLogger("user_level_log").info(
            f"Data path: {data_root_path}{fname_prefix}*.h5"
        )

        try:
            HWR.beamline.diffractometer.start_ssx_scan(params.sub_sampling)
            HWR.beamline.diffractometer.wait_ready()
        except:
            msg = "Diffractometer failed! Waiting for detector to finish"
            logging.getLogger("user_level_log").error(msg)
            HWR.beamline.detector.wait_ready()
            raise

        if HWR.beamline.control.safshut_oh2.state.name == "OPEN":
            HWR.beamline.control.safshut_oh2.close()
            logging.getLogger("user_level_log").info("shutter closed")

        HWR.beamline.detector.wait_ready()
        logging.getLogger("user_level_log").info(f"Acquired {region}")

    def pre_execute(self):
        super().pre_execute()

    def post_execute(self):
        super().post_execute()

    def stop(self):
        super().stop()



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""EMBLCRL"""

import logging
import math

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class EMBLCRL(HardwareObject):
    """Controls CRLs"""

    def __init__(self, name):

        HardwareObject.__init__(self, name)

        self.focal_length = None
        self.lens_count = None

        self.modes = None
        self.current_mode = None
        self.energy_value = None
        self.energy_state = None
        self.current_focusing_mode = None
        self.crl_value = None
        self.chan_crl_value = None
        self.cmd_set_crl_value = None
        self.cmd_set_trans_value = None

        self.beam_focusing_hwobj = None

        self.at_startup = None

    def init(self):
        """Inits all variables"""

        self.focal_length = self.get_property("focal_length")
        self.lens_count = 6

        self.chan_crl_value = self.get_channel_object("chanCrlValue")
        if self.chan_crl_value:
            self.chan_crl_value.connect_signal("update", self.crl_value_changed)

        self.cmd_set_crl_value = self.get_command_object("cmdSetLenses")
        self.cmd_set_trans_value = self.get_command_object("cmdSetTrans")

        self.energy_value = HWR.beamline.energy.get_value()
        self.connect(HWR.beamline.energy, "stateChanged", self.energy_state_changed)

        self.beam_focusing_hwobj = self.get_object_by_role("beam_focusing")
        if self.beam_focusing_hwobj is not None:
            self.connect(
                self.beam_focusing_hwobj,
                "focusingModeRequested",
                self.focusing_mode_requested,
            )

            lens_modes = self.beam_focusing_hwobj.get_available_lens_modes()
            if lens_modes:
                self.set_mode(lens_modes[0])
        else:
            self.set_mode("Manual")

    def convert_value(self, value):
        """Converts int to list of bits"""
        if isinstance(value, (list, tuple)):
            lens_combination = 0
            for index in range(self.lens_count):
                lens_combination = lens_combination + value[index] * pow(2, index)
        else:
            lens_combination = [0, 0, 0, 0, 0, 0]
            for index in range(self.lens_count):
                lens_combination[index] = (value & pow(2, index)) / pow(2, index)
        return lens_combination

    def get_modes(self):
        """Returns list with available CRL modes"""
        if self.beam_focusing_hwobj:
            return self.beam_focusing_hwobj.get_available_lens_modes()
        else:
            return ["Manual"]

    def get_mode(self):
        """Returns current crl mode"""
        return self.current_mode

    def set_mode(self, mode):
        """Sets crl mode

        :param mode: crl mode
        :type mode: str
        :return: None
        """
        self.current_mode = mode
        # if self.current_mode == "Out":
        #    self.set_crl_value([0, 0, 0, 0, 0, 0])
        if self.current_mode == "Automatic":
            self.set_according_to_energy()
        self.emit("crlModeChanged", self.current_mode)

    def energy_state_changed(self, state):
        """If CRL's in the automatic mode then change setting accoring
           to the current energy

        :param state: energy state
        :type state: str
        """
        if (
            state == "ready"
            and state != self.energy_state
            and self.current_mode == "Automatic"
        ):
            self.energy_value = HWR.beamline.energy.get_value()
            self.set_according_to_energy()
        self.energy_state = state

    def set_according_to_energy(self):
        """Sets CRL combination according to the current energy"""
        min_abs = 20
        selected_combination = None
        # crl_value = [0, 0, 0, 0, 0, 0]

        self.energy_value = HWR.beamline.energy.get_value()
        for combination_index in range(1, 65):
            current_abs = abs(
                self.energy_value
                - math.sqrt(
                    (2 * 341.52 * combination_index)
                    / (2000 * (1 / 42.6696 + 1 / self.focal_length))
                )
            )
            if current_abs < min_abs:
                min_abs = current_abs
                selected_combination = combination_index
        # for index in range(6):
        #    crl_value[index] = (selected_combination & pow(2,index))/pow(2,index)
        self.set_crl_value(self.convert_value(selected_combination))

    def get_image_plane_distance(self, value):
        """Returns image plane distance"""
        if isinstance(value, list):
            lens_combination = self.convert_value(value)
        else:
            lens_combination = value
        return 1.0 / (
            2 * 341.52 * lens_combination / 2000 / (self.energy_value ** 2)
            - 1 / 42.6696
        )

    def focusing_mode_requested(self, focusing_mode):
        """Sets CRL combination based on the focusing mode"""
        if focusing_mode is not None:
            self.modes = self.beam_focusing_hwobj.get_available_lens_modes(
                focusing_mode
            )
            self.set_mode(self.modes[0])
            self.set_crl_value(
                self.beam_focusing_hwobj.get_lens_combination(focusing_mode)
            )

    def crl_value_changed(self, value):
        """Emit signal when crl combination changed"""
        self.crl_value = value
        self.emit("crlValueChanged", self.crl_value)

    def set_crl_value(self, value, timeout=None):
        """Sets CRL lens combination. If integer passed then
           converts value to the bit list
        """
        if not isinstance(value, (list, tuple)):
            value = self.convert_value(value)

        if value is not None:
            self.cmd_set_crl_value(value)
            self.cmd_set_trans_value(1)
            logging.getLogger("GUI").info(
                "Setting CRL image plane "
                + "distance to %.2f" % self.get_image_plane_distance(value)
            )

            if timeout:
                gevent.sleep(1)
                with gevent.Timeout(timeout, Exception("Timeout waiting for CRL")):
                    while value != self.crl_value:
                        gevent.sleep(0.1)

    def get_crl_value(self):
        """Return crl combination"""
        return self.crl_value

    def re_emit_values(self):
        """Reemits signals"""
        self.emit("crlModeChanged", self.current_mode)
        self.emit("crlValueChanged", self.crl_value)

    def move_up(self):
        """Moves lense combination one value up"""
        new_value = self.convert_value(self.crl_value) + 1
        self.set_crl_value(new_value)

    def move_down(self):
        """Moves lense combination one value down"""
        new_value = self.convert_value(self.crl_value) - 1
        self.set_crl_value(new_value)



import json
import logging
import pathlib
import shutil
from datetime import (
    datetime,
    timedelta,
)
from typing import List

from pyicat_plus.client.main import (
    IcatClient,
    IcatInvestigationClient,
)
from pyicat_plus.client.models.session import Session as ICATSession

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.HardwareObjects.abstract.AbstractLims import AbstractLims
from mxcubecore.model.lims_session import (
    Lims,
    LimsSessionManager,
    Session,
)


class ICATLIMS(AbstractLims):
    """
    ICAT client.
    """

    def __init__(self, name):
        super().__init__(name)
        HardwareObject.__init__(self, name)
        self.investigations = None
        self.icatClient = None
        self.catalogue = None
        self.lims_rest = None
        self.ingesters = None

    def init(self):
        self.url = self.get_property("ws_root")
        self.ingesters = self.get_property("queue_urls")
        self.investigations = []

        # Initialize ICAT client and catalogue
        self.icatClient = IcatClient(
            catalogue_url=self.url,
            tracking_url=self.url,
            metadata_urls=["bcu-mq-01:61613"],
            catalogue_queues=["bcu-mq-01:61613"],
        )

    def get_lims_name(self) -> List[Lims]:
        return [
            Lims(name="DRAC", description="Data Repository for Advancing open sCience"),
        ]

    def login(
        self, user_name: str, password: str, is_local_host: bool
    ) -> LimsSessionManager:

        logging.getLogger("HWR").debug("[ICAT] authenticate %s" % (user_name))

        self.icat_session: ICATSession = self.icatClient.do_log_in(password)
        # Catalogue client to retrieve investigations
        self.catalogue = self.icatClient.catalogue_client

        # Catalogue client to retrieve parcels and samples
        self.tracking = self.icatClient.tracking_client

        if self.catalogue is None or self.tracking is None:
            logging.getLogger("HWR").error(
                "[ICAT] Error initializing catalogue/tracking. catalogue=%s tracking=%s"
                % (self.url, self.url)
            )
            raise RuntimeError("Could not initialize catalogue/tracking")

        # Connected to metadata catalogue
        logging.getLogger("HWR").debug(
            "[ICAT] Connected succesfully to catalogue. fullName=%s url=%s"
            % (self.icat_session["fullName"], self.url)
        )

        # Retrieving user's investigations
        sessions = self.to_sessions(self.__get_all_investigations())
        logging.getLogger("HWR").debug(
            "[ICAT] Successfully retrieved %s sessions" % (len(sessions))
        )
        self.set_sessions(sessions)
        return self.session_manager

    def is_user_login_type(self) -> bool:
        return True

    def get_proposals_by_user(self, user_name):
        logging.getLogger("HWR").debug("get_proposals_by_user %s" % user_name)

        logging.getLogger("HWR").debug(
            "[ICATCLient] Read %s investigations" % len(self.lims_rest.investigations)
        )
        return self.lims_rest.to_sessions(self.lims_rest.investigations)

    def get_samples(self, lims_name):
        try:
            logging.getLogger("HWR").debug(
                "[ICATClient] get_samples %s %s lims_name=%s",
                self.session_manager.active_session.session_id,
                self.session_manager.active_session.proposal_name,
                lims_name,
            )
            parcels = self.get_parcels_by_investigation_id()
            queue_samples = []
            for parcel in parcels:
                pucks = parcel["content"]
                logging.getLogger("HWR").debug(
                    "[ICATClient] Reading parcel '%s' with '%s' pucks"
                    % (parcel["name"], len(pucks))
                )
                # Parcels contains pucks: unipucks and spine pucks
                for puck in pucks:
                    tracking_samples = puck["content"]
                    if "sampleChangerLocation" in puck:
                        logging.getLogger("HWR").debug(
                            "[ICATClient] Processing puck '%s' within parcel '%s' at position '%s'. Number of samples '%s'"
                            % (
                                puck["name"],
                                parcel["name"],
                                puck["sampleChangerLocation"],
                                len(tracking_samples),
                            )
                        )
                        for tracking_sample in tracking_samples:
                            queue_samples.append(
                                self.__to_sample(tracking_sample, puck)
                            )

        except Exception as e:
            logging.getLogger("HWR").error(e)
            return []

        logging.getLogger("HWR").debug(
            "[ICATClient] Read %s samples" % (len(queue_samples))
        )

        return queue_samples

    def find(self, arr, atribute_name):
        for x in arr:
            if x["key"] == atribute_name:
                return x["value"]
        return ""

    def __to_sample(self, tracking_sample, puck):
        """Converts the sample tracking into the expected sample data structure"""
        experiment_plan = tracking_sample["experimentPlan"]
        return {
            "cellA": self.find(experiment_plan, "unit_cell_a"),
            "cellAlpha": self.find(experiment_plan, "unit_cell_alpha"),
            "cellB": self.find(experiment_plan, "unit_cell_b"),
            "cellBeta": self.find(experiment_plan, "unit_cell_beta"),
            "cellC": self.find(experiment_plan, "unit_cell_c"),
            "cellGamma": self.find(experiment_plan, "unit_cell_gamma"),
            "containerSampleChangerLocation": str(puck["sampleChangerLocation"]),
            "crystalSpaceGroup": self.find(experiment_plan, "forceSpaceGroup"),
            "diffractionPlan": {
                # "diffractionPlanId": 457980, TODO: do we need this?
                "experimentKind": self.find(experiment_plan, "experimentKind"),
                "numberOfPositions": self.find(experiment_plan, "numberOfPositions"),
                "observedResolution": self.find(experiment_plan, "observedResolution"),
                "preferredBeamDiameter": self.find(
                    experiment_plan, "preferredBeamDiameter"
                ),
                "radiationSensitivity": self.find(
                    experiment_plan, "radiationSensitivity"
                ),
                "requiredCompleteness": self.find(
                    experiment_plan, "requiredCompleteness"
                ),
                "requiredMultiplicity": self.find(
                    experiment_plan, "requiredMultiplicity"
                ),
                "requiredResolution": self.find(experiment_plan, "requiredResolution"),
            },
            "experimentType": self.find(experiment_plan, "workflowType"),
            "proteinAcronym": tracking_sample["name"],
            "sampleId": tracking_sample["sampleId"],
            "sampleLocation": tracking_sample["sampleContainerPosition"],
            "sampleName": str(tracking_sample["name"]),
            "smiles": None,
        }

    def create_session(self, session_dict):
        pass

    def _store_data_collection_group(self, group_data):
        pass

    def store_robot_action(self, proposal_id: str):
        raise Exception("Not implemented")

    @property
    def filter(self):
        return self.get_property("filter", None)

    @property
    def override_beamline_name(self):
        return self.get_property(
            "override_beamline_name", HWR.beamline.session.beamline_name
        )

    @property
    def compatible_beamlines(self):
        return self.get_property(
            "compatible_beamlines", HWR.beamline.session.beamline_name
        )

    @property
    def data_portal_url(self):
        return self.get_property("data_portal_url", None)

    @property
    def user_portal_url(self):
        return self.get_property("user_portal_url", None)

    @property
    def logbook_url(self):
        return self.get_property("logbook_url", None)

    @property
    def before_offset_days(self):
        return self.get_property("before_offset_days", "1")

    @property
    def after_offset_days(self):
        return self.get_property("after_offset_days", "1")

    def _string_to_format_date(self, date: str, format: str) -> str:
        if date is not None:
            date_time = self._tz_aware_fromisoformat(date)
            if date_time is not None:
                return date_time.strftime(format)
        return ""

    def _string_to_date(self, date: str) -> str:
        return self._string_to_format_date(date, "%Y%m%d")

    def _string_to_time(self, date: str) -> str:
        return self._string_to_format_date(date, "%H:%M:%S")

    def _tz_aware_fromisoformat(self, date: str) -> datetime:
        try:
            return datetime.fromisoformat(date).astimezone()
        except Exception:
            return None

    def set_active_session_by_id(self, session_id: str) -> Session:

        if self.is_session_already_active(self.session_manager.active_session):
            return self.session_manager.active_session

        sessions = self.session_manager.sessions

        if len(sessions) == 0:
            logging.getLogger("HWR").error(
                "Session list is empty. No session candidates"
            )
            raise Exception("No sessions available")

        if len(sessions) == 1:
            self.session_manager.active_session = sessions[0]
            logging.getLogger("HWR").debug(
                "Session list contains a single session. proposal_name=%s",
                self.session_manager.active_session.proposal_name,
            )
            return self.session_manager.active_session

        session_list = [obj for obj in sessions if obj.session_id == session_id]
        if len(session_list) != 1:
            raise Exception(
                "Session not found in the local list of sessions. session_id="
                + session_id
            )
        self.session_manager.active_session = session_list[0]
        return self.session_manager.active_session

    def allow_session(self, session: Session):
        self.active_session = session
        logging.getLogger("HWR").debug(
            "[ICAT] allow_session investigationId=%s", session.session_id
        )
        self.catalogue.reschedule_investigation(session.session_id)

    def get_session_by_id(self, id: str):
        logging.getLogger("HWR").debug(
            "[ICAT] get_session_by_id investigationId=%s investigations=%s",
            id,
            str(len(self.investigations)),
        )
        investigation_list = list(filter(lambda p: p["id"] == id, self.investigations))
        if len(investigation_list) == 1:
            self.investigation = investigation_list[0]
            return self.__to_session(investigation_list[0])
        logging.getLogger("HWR").warn(
            "[ICAT] No investigation found. get_session_by_id investigationId=%s investigations=%s",
            id,
            str(len(self.investigations)),
        )
        return None

    def __get_all_investigations(self):
        """Returns all investigations by user. An investigation corresponds to
        one experimental session. It returns an empty array in case of error"""
        try:
            self.investigations = []
            logging.getLogger("HWR").debug(
                "[ICAT] __get_all_investigations before=%s after=%s beamline=%s isInstrumentScientist=%s isAdministrator=%s compatible_beamlines=%s"
                % (
                    self.before_offset_days,
                    self.after_offset_days,
                    self.override_beamline_name,
                    self.icat_session["isInstrumentScientist"],
                    self.icat_session["isAdministrator"],
                    self.compatible_beamlines,
                )
            )

            if self.icat_session is not None and (
                self.icat_session["isAdministrator"]
                or self.icat_session["isInstrumentScientist"]
            ):
                self.investigations = self.catalogue.get_investigations_by(
                    start_date=datetime.today()
                    - timedelta(days=float(self.before_offset_days)),
                    end_date=datetime.today()
                    + timedelta(days=float(self.after_offset_days)),
                    instrument_name=self.compatible_beamlines,
                )
            else:
                self.investigations = self.catalogue.get_investigations_by(
                    filter=self.filter,
                    instrument_name=self.override_beamline_name,
                    start_date=datetime.today()
                    - timedelta(days=float(self.before_offset_days)),
                    end_date=datetime.today()
                    + timedelta(days=float(self.after_offset_days)),
                )
            logging.getLogger("HWR").debug(
                "[ICAT] __get_all_investigations retrieved %s investigations"
                % len(self.investigations)
            )
            return self.investigations
        except Exception as e:
            self.investigations = []
            logging.getLogger("HWR").error("[ICAT] __get_all_investigations %s " % e)
        return self.investigations

    def __get_proposal_number_by_investigation(self, investigation):
        """
        Given an investigation it returns the proposal number.
        Example: investigation["name"] = "MX-1234"
        returns: 1234

        TODO: this might not work for all type of proposals (example: TEST proposals)
        """
        return (
            investigation["name"]
            .replace(investigation["type"]["name"], "")
            .replace("-", "")
        )

    def _get_data_portal_url(self, investigation):
        try:
            return (
                self.data_portal_url.replace("{id}", str(investigation["id"]))
                if self.data_portal_url is not None
                else ""
            )
        except Exception:
            return ""

    def _get_logbook_url(self, investigation):
        try:
            return (
                self.logbook_url.replace("{id}", str(investigation["id"]))
                if self.logbook_url is not None
                else ""
            )
        except Exception:
            return ""

    def _get_user_portal_url(self, investigation):
        try:
            return (
                self.user_portal_url.replace(
                    "{id}", str(investigation["parameters"]["Id"])
                )
                if self.user_portal_url is not None
                and investigation["parameters"]["Id"] is not None
                else ""
            )
        except Exception:
            return ""

    def __to_session(self, investigation) -> Session:
        """This methods converts a ICAT investigation into a session"""

        actual_start_date = (
            investigation["parameters"]["__actualStartDate"]
            if "__actualStartDate" in investigation["parameters"]
            else investigation["startDate"]
        )
        actual_end_date = (
            investigation["parameters"]["__actualEndDate"]
            if "__actualEndDate" in investigation["parameters"]
            else investigation.get("endDate", None)
        )
        # If session has been rescheduled new date is overwritten
        return Session(
            code=investigation["type"]["name"],
            number=self.__get_proposal_number_by_investigation(investigation),
            title=f'{investigation["title"]}',
            session_id=investigation["id"],
            proposal_id=investigation["id"],
            proposal_name=investigation["name"],
            beamline_name=investigation["instrument"]["name"],
            comments="",
            start_datetime=investigation.get(
                "startDate", None
            ),  # self._string_to_date(investigation.get("startDate", None)),
            start_date=self._string_to_date(investigation.get("startDate", None)),
            start_time=self._string_to_time(investigation.get("startDate", None)),
            end_datetime=investigation.get("endDate", None),
            end_date=self._string_to_date(
                investigation.get("endDate", None)
            ),  # self._string_to_time(investigation.get("endDate", None)),
            end_time=self._string_to_time(investigation.get("endDate", None)),
            actual_start_date=self._string_to_date(actual_start_date),
            actual_start_time=self._string_to_time(actual_start_date),
            actual_end_date=self._string_to_date(actual_end_date),
            actual_end_time=self._string_to_time(actual_end_date),
            nb_shifts=3,
            scheduled=self.is_scheduled_on_host_beamline(
                investigation["instrument"]["name"]
            ),
            is_scheduled_time=self.is_scheduled_now(actual_start_date, actual_end_date),
            is_scheduled_beamline=self.is_scheduled_on_host_beamline(
                investigation["instrument"]["name"]
            ),
            data_portal_URL=self._get_data_portal_url(investigation),
            user_portal_URL=self._get_user_portal_url(investigation),
            logbook_URL=self._get_logbook_url(investigation),
            is_rescheduled=(
                True if "__actualEndDate" in investigation["parameters"] else False
            ),
        )

    def get_full_user_name(self):
        return self.icat_session["fullName"]

    def get_user_name(self):
        return self.icat_session["username"]

    def to_sessions(self, investigations):
        return [self.__to_session(investigation) for investigation in investigations]

    def get_parcels_by_investigation_id(self):
        """Returns the parcels associated to an investigation"""
        try:
            logging.getLogger("HWR").debug(
                "[ICAT] Retrieving parcels by investigation_id %s "
                % (self.session_manager.active_session.session_id)
            )
            parcels = self.tracking.get_parcels_by(
                self.session_manager.active_session.session_id
            )
            logging.getLogger("HWR").debug(
                "[ICAT] Successfully retrieved %s parcels" % (len(parcels))
            )
            return parcels
        except Exception as e:
            logging.getLogger("HWR").error(
                "[ICAT] get_parcels_by_investigation_id %s " % (str(e))
            )
        return []

    def echo(self):
        """Mockup for the echo method."""
        return True

    def is_connected(self):
        return self.login_ok

    def __add_protein_acronym(self, sample_node, metadata):
        """
        Fills the sample acronym that should match with the acronym defined in the sample sheet
        """
        if sample_node is not None:
            if sample_node.crystals is not None:
                if len(sample_node.crystals) > 0:
                    crystal = sample_node.crystals[0]
                    if crystal.protein_acronym is not None:
                        metadata["SampleProtein_acronym"] = crystal.protein_acronym

    def __add_sample_changer_position(self, cell, puck, metadata):
        """
        Adds to the sample changer position based on the cell and the puck number

        Args:
            cell(str): cell position of the puck in the sample changer
            puck(str): position of the puck within the cell
            metadata(dict): metadata to be pushed to ICAT
        """
        try:
            if cell is not None and puck is not None:
                position = int(cell * 3) + int(puck)
                metadata["SampleChanger_position"] = position
        except Exception as e:
            logging.getLogger("HWR").exception(e)

    def add_sample_metadata(self, metadata, collection_parameters):
        """
        Adds to the metadata dictionary the metadata concerning sample position, container and tracking

        Args:
            metadata(dict): metadata to be pushed to ICAT
            collection_parameters(dict): Data collection parameters
        """
        try:
            queue_entry = HWR.beamline.queue_manager.get_current_entry()
            sample_node = queue_entry.get_data_model().get_sample_node()
            # sample_node.name this is name of the sample

            location = sample_node.location  # Example: (8,2,5)

            if len(location) == 3:
                (cell, puck, sample_position) = location
            else:
                cell = 1
                (puck, sample_position) = location

            self.__add_sample_changer_position(cell, puck, metadata)
            metadata["SampleTrackingContainer_position"] = sample_position
            metadata["SampleTrackingContainer_type"] = (
                "UNIPUCK"  # this could be read from the configuration file somehow
            )
            metadata["SampleTrackingContainer_capaticy"] = (
                "16"  # this could be read from the configuration file somehow
            )

            self.__add_protein_acronym(sample_node, metadata)

            if HWR.beamline.lims is not None:
                sample = HWR.beamline.lims.find_sample_by_sample_id(
                    collection_parameters.get("blSampleId")
                )
                if sample is not None:
                    if "containerCode" in sample:
                        metadata["SampleTrackingContainer_id"] = sample["containerCode"]
                    else:
                        metadata["SampleTrackingContainer_id"] = (
                            str(cell) + "_" + str(puck)
                        )  # Fake identifier that needs to be replaced by container code

        except Exception as e:
            logging.getLogger("HWR").exception(e)

    def store_beamline_setup(self, session_id: str, bl_config_dict: dict):
        pass

    def store_image(self, image_dict: dict):
        pass

    def store_energy_scan(self, energyscan_dict: dict):
        pass

    def store_xfe_spectrum(self, xfespectrum_dict: dict):
        pass

    def store_workflow(self, workflow_dict: dict):
        pass

    def store_data_collection(self, mx_collection, bl_config=None):
        pass

    def update_data_collection(self, mx_collection):
        pass

    def finalize_data_collection(self, collection_parameters):
        logging.getLogger("HWR").info("Storing datacollection in ICAT")
        if event == "END":
            try:
                fileinfo = collection_parameters["fileinfo"]
                directory = pathlib.Path(fileinfo["directory"])
                dataset_name = directory.name
                # Determine the scan type
                if dataset_name.endswith("mesh"):
                    scanType = "mesh"
                elif dataset_name.endswith("line"):
                    scanType = "line"
                elif dataset_name.endswith("characterisation"):
                    scanType = "characterisation"
                elif dataset_name.endswith("datacollection"):
                    scanType = "datacollection"
                else:
                    scanType = collection_parameters["experiment_type"]
                workflow_params = collection_parameters.get("workflow_parameters", {})
                workflow_type = workflow_params.get("workflow_type")
                if workflow_type is None:
                    if not directory.name.startswith("run"):
                        dataset_name = fileinfo["prefix"]

                if collection_parameters["sample_reference"]["acronym"]:
                    sample_name = (
                        collection_parameters["sample_reference"]["acronym"]
                        + "-"
                        + collection_parameters["sample_reference"]["sample_name"]
                    )
                else:
                    sample_name = collection_parameters["sample_reference"][
                        "sample_name"
                    ].replace(":", "-")

                logging.getLogger("HWR").info(f"LIMS sample name {sample_name}")
                oscillation_sequence = collection_parameters["oscillation_sequence"][0]
                beamline = HWR.beamline.session.beamline_name.lower()
                distance = HWR.beamline.detector.distance.get_value()
                proposal = f"{HWR.beamline.session.proposal_code}{HWR.beamline.session.proposal_number}"
                metadata = {
                    "MX_beamShape": collection_parameters["beamShape"],
                    "MX_beamSizeAtSampleX": collection_parameters["beamSizeAtSampleX"],
                    "MX_beamSizeAtSampleY": collection_parameters["beamSizeAtSampleY"],
                    "MX_dataCollectionId": collection_parameters["collection_id"],
                    "MX_detectorDistance": distance,
                    "MX_directory": str(directory),
                    "MX_exposureTime": oscillation_sequence["exposure_time"],
                    "MX_flux": collection_parameters["flux"],
                    "MX_fluxEnd": collection_parameters["flux_end"],
                    "MX_positionName": collection_parameters["position_name"],
                    "MX_numberOfImages": oscillation_sequence["number_of_images"],
                    "MX_oscillationRange": oscillation_sequence["range"],
                    "MX_oscillationStart": oscillation_sequence["start"],
                    "MX_oscillationOverlap": oscillation_sequence["overlap"],
                    "MX_resolution": collection_parameters["resolution"],
                    "scanType": scanType,
                    "MX_startImageNumber": oscillation_sequence["start_image_number"],
                    "MX_template": fileinfo["template"],
                    "MX_transmission": collection_parameters["transmission"],
                    "MX_xBeam": collection_parameters["xBeam"],
                    "MX_yBeam": collection_parameters["yBeam"],
                    "Sample_name": sample_name,
                    "InstrumentMonochromator_wavelength": collection_parameters[
                        "wavelength"
                    ],
                    "Workflow_name": workflow_params.get("workflow_name"),
                    "Workflow_type": workflow_params.get("workflow_type"),
                    "Workflow_id": workflow_params.get("workflow_uid"),
                    "MX_kappa_settings_id": workflow_params.get(
                        "workflow_kappa_settings_id"
                    ),
                    "MX_characterisation_id": workflow_params.get(
                        "workflow_characterisation_id"
                    ),
                    "MX_position_id": workflow_params.get("workflow_position_id"),
                    "group_by": workflow_params.get("workflow_group_by"),
                }
                # Store metadata on disk
                self.add_sample_metadata(metadata, collection_parameters)
                icat_metadata_path = pathlib.Path(directory) / "metadata.json"
                with open(icat_metadata_path, "w") as f:
                    f.write(json.dumps(metadata, indent=4))
                # Create ICAT gallery
                gallery_path = directory / "gallery"
                gallery_path.mkdir(mode=0o755, exist_ok=True)
                for snapshot_index in range(1, 5):
                    key = f"xtalSnapshotFullPath{snapshot_index}"
                    if key in collection_parameters:
                        snapshot_path = pathlib.Path(collection_parameters[key])
                        if snapshot_path.exists():
                            logging.getLogger("HWR").debug(
                                f"Copying snapshot index {snapshot_index} to gallery"
                            )
                            shutil.copy(snapshot_path, gallery_path)
                logging.getLogger("HWR").info(f"Beamline: {beamline}")
                logging.getLogger("HWR").info(f"Proposal: {proposal}")

                self.icatClient.store_dataset(
                    beamline=beamline,
                    proposal=proposal,
                    dataset=dataset_name,
                    path=str(directory),
                    metadata=metadata,
                )
                logging.getLogger("HWR").debug("Done uploading to ICAT")
            except Exception as e:
                logging.getLogger("HWR").exception(e)

    def update_bl_sample(self, bl_sample: str):
        """
        Creates or stos a BLSample entry.
        # NBNB update doc string
        :param sample_dict: A dictonary with the properties for the entry.
        :type sample_dict: dict
        """
        pass



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.


"""
EMBLOnlineProcessing
"""

import ast
import logging
import os
import subprocess

import gevent
import numpy as np

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractOnlineProcessing import (
    AbstractOnlineProcessing,
)
from mxcubecore.HardwareObjects.XSDataCommon import (
    XSDataBoolean,
    XSDataDouble,
    XSDataInteger,
    XSDataString,
)
from mxcubecore.HardwareObjects.XSDataControlDozorv1_1 import (
    XSDataControlImageDozor,
    XSDataInputControlDozor,
    XSDataResultControlDozor,
)

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"


CRYSTFEL_GEOM_FILE_TEMPLATE = """clen = {:.5f}
photon_energy = {:.1f}

adu_per_photon = 1
res = {:.1f}

panel0/min_fs = 0
panel0/min_ss = 0
panel0/max_fs = {:.0f}
panel0/max_ss = {:.0f}
panel0/corner_x = {:.2f}
panel0/corner_y = {:.2f}
panel0/fs = x
panel0/ss = y

badBeam/min_x = -10.0
badBeam/max_x = +10.0
badBeam/min_y = -10.0
badBeam/max_y = +10.0
"""

CRYSTFEL_CELL_FILE_TEMPLATE = """CrystFEL unit cell file version 1.0

lattice_type = {:s}
centering = {:s}
{:s}

a = {:.2f} A
b = {:.2f} A
c = {:.2f} A
al = {:.2f} deg
be = {:.2f} deg
ga = {:.2f} deg
"""

CONST_H = 4.135667516e-15  # eV*s
CONST_C = 299792458.0


class EMBLOnlineProcessing(AbstractOnlineProcessing):
    """
    Obtains Dozor on the fly processing results
    Assembles crystfel input files and starts crystfel
    """

    def __init__(self, name):
        AbstractOnlineProcessing.__init__(self, name)

        self.display_task = None
        self.nxds_input_template = None

        self.crystfel_script = None
        self.crystfel_script_template = None
        self.crystfel_params = None

        self.all_frames_batch_processed = False
        self.all_frames_dozor_is = False

        self.is_count = 0
        self.batch_count = 0

        self.chan_dozor_pass = None
        self.chan_dozor_average_i = None
        self.chan_frame_count = None

        self.result_types = []

    def init(self):
        self.chan_dozor_average_i = self.get_channel_object("chanDozorAverageI")
        if self.chan_dozor_average_i is not None:
            key = "average_intensity"
            self.result_types.append(
                {"key": key, "descr": "Average I", "color": (255, 0, 0), "size": 0}
            )
            self.chan_dozor_average_i.connect_signal(
                "update", self.dozor_average_i_changed
            )

        AbstractOnlineProcessing.init(self)

        self.chan_dozor_pass = self.get_channel_object("chanDozorPass")
        if self.chan_dozor_pass is not None:
            self.chan_dozor_pass.connect_signal("update", self.batch_processed)

        self.chan_frame_count = self.get_channel_object("chanFrameCount")
        if self.chan_frame_count is not None:
            self.chan_frame_count.connect_signal("update", self.frame_count_changed)

        self.chan_dozor_is = self.get_channel_object("chanDozorIS")
        if self.chan_dozor_is is not None:
            self.result_types.append(
                {
                    "key": "is",
                    "descr": "Intensity",
                    "color": (255, 0, 0)
                }
            )
            self.chan_dozor_is.connect_signal("update", self.dozor_is_changed)

        self.crystfel_script = self.get_property("crystfel_script")

        if self.get_property("nxds_input_template_file") is not None:
            with open(
                self.get_property("nxds_input_template_file"), "r"
            ) as template_file:
                self.nxds_input_template = "".join(template_file.readlines())

        if self.get_property("crystfel_script_template_file") is not None:
            with open(
                self.get_property("crystfel_script_template_file"), "r"
            ) as template_file:
                self.crystfel_script_template = "".join(template_file.readlines())

        if self.get_property("crystfel_params") is not None:
            self.crystfel_params = ast.literal_eval(
                self.get_property("crystfel_params")
            )

    def create_processing_input_file(self, processing_input_filename):
        """Creates dozor input file base on data collection parameters

        :param processing_input_filename
        :type : str
        """
        input_file = XSDataInputControlDozor()
        input_file.setTemplate(XSDataString(self.params_dict["template"]))
        input_file.setFirst_image_number(
            XSDataInteger(self.params_dict["first_image_num"])
        )
        input_file.setLast_image_number(XSDataInteger(self.params_dict["images_num"]))
        input_file.setFirst_run_number(XSDataInteger(self.params_dict["run_number"]))
        input_file.setLast_run_number(XSDataInteger(self.params_dict["run_number"]))
        input_file.setLine_number_of(XSDataInteger(self.params_dict["lines_num"]))
        input_file.setReversing_rotation(
            XSDataBoolean(self.params_dict["reversing_rotation"])
        )
        input_file.setPixelMin(XSDataInteger(HWR.beamline.detector.get_pixel_min()))
        input_file.setPixelMax(XSDataInteger(HWR.beamline.detector.get_pixel_max()))
        input_file.setBeamstopSize(XSDataDouble(HWR.beamline.beamstop.get_size()))
        input_file.setBeamstopDistance(XSDataDouble(HWR.beamline.beamstop.get_distance()))
        input_file.setBeamstopDirection(
            XSDataString(HWR.beamline.beamstop.get_direction())
        )

        input_file.exportToFile(processing_input_filename)

    def run_processing(self, data_collection):
        """
        :param data_collection: data collection object
        :type data_collection: queue_model_objects.DataCollection
        """
        self.data_collection = data_collection
        self.prepare_processing()

        if self.crystfel_script:
            all_file_filename = self.create_all_file_list()
            self.start_crystfel_autoproc(all_file_filename)

        self.started = True
        self.display_task = gevent.spawn(self.update_map)

        if self.chan_dozor_pass is None:
            # Start dozor via EDNA

            input_filename = os.path.join(
                    self.params_dict["process_directory"],
                    "dozor_input.xml"
            )
            self.create_processing_input_file(input_filename)

            if not os.path.isfile(self.start_command):
                msg = (
                    "OnlineProcessing: Start command %s" % self.start_command
                    + "is not executable"
                )
                logging.getLogger("queue_exec").error(msg)
                self.set_processing_status("Failed")
            else:
                input_filename = os.path.join(
                    self.params_dict["process_directory"],
                    "dozor_input.xml"
                )
                self.create_processing_input_file(input_filename)

                line_to_execute = (
                    self.start_command
                    + " "
                    + input_filename
                    + " "
                    + self.params_dict["process_directory"]
                )

                self.started = True
                subprocess.Popen(
                    str(line_to_execute),
                    shell=True,
                    stdin=None,
                    stdout=None,
                    stderr=None,
                    close_fds=True,
                )

        self.emit(
            "processingStarted",
            (self.params_dict, self.results_raw, self.results_aligned),
        )
        self.emit("processingResultsUpdate", False)

    def frame_count_changed(self, frame_count):
        """
        Finishes processing if the last frame is processed
        :param frame_count:
        :return:
        """
        if isinstance(frame_count, tuple):
            frame_count=frame_count[1]
        if self.started:
            self.emit("processingFrame", frame_count)
            """
            if frame_count >= self.params_dict["images_num"] - 1:
                print "done!"
                self.set_processing_status("Success")
            """

    def smooth(self):
        """
        Smooths the resolution
        :return:
        """
        step = 30
        for key in ["score", "spots_num"]:
            for index in range(self.results_raw[key].size):
                self.results_raw[key][index] = numpy.mean(
                    self.results_raw[key][index - step : index + step]
                )

    def get_spacegroup_dict(self, proc_params, space_group_name):
        spacegroup_dict = {}

        if "P212121" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "orthorhombic"
            spacegroup_dict["point_group"]        = "mmm"
            spacegroup_dict["space_group"]        = "P212121"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = ""
            spacegroup_dict["space_group_number"] = 19
        elif "P21212" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "orthorhombic"
            spacegroup_dict["point_group"]        = "mmm"
            spacegroup_dict["space_group"]        = "P21212"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = ""
            spacegroup_dict["space_group_number"] = 18
        elif "P213" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "cubic"
            spacegroup_dict["point_group"]        = "m-3"
            spacegroup_dict["space_group"]        = "P213"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = ""
            spacegroup_dict["space_group_number"] = 198
        elif "P21" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "monoclinic"
            spacegroup_dict["point_group"]        = "2/m"
            spacegroup_dict["space_group"]        = "P21"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = ""
            spacegroup_dict["space_group_number"] = 4
        elif "P3112" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "hexagonal"
            spacegroup_dict["point_group"]        = "-31m_H"
            spacegroup_dict["space_group"]        = "P3112"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = "unique_axis = c"
            spacegroup_dict["space_group_number"] = 151
        elif "P3121" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "hexagonal"
            spacegroup_dict["point_group"]        = "-3m1_H"
            spacegroup_dict["space_group"]        = "P3121"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = "unique_axis = c"
            spacegroup_dict["space_group_number"] = 152
        elif "P3212" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "hexagonal"
            spacegroup_dict["point_group"]        = "-31m_H"
            spacegroup_dict["space_group"]        = "P3212"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = "unique_axis = c"
            spacegroup_dict["space_group_number"] = 153
        elif "P3221" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "hexagonal"
            spacegroup_dict["point_group"]        = "-3m1_H"
            spacegroup_dict["space_group"]        = "P3221"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = "unique_axis = c"
            spacegroup_dict["space_group_number"] = 154
        elif "P321" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "hexagonal"
            spacegroup_dict["point_group"]        = "-3m1_H"
            spacegroup_dict["space_group"]        = "P321"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = "unique_axis = c"
            spacegroup_dict["space_group_number"] = 150
        elif "P32" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "hexagonal"
            spacegroup_dict["point_group"]        = "-3_H"
            spacegroup_dict["space_group"]        = "P32"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = "unique_axis = c"
            spacegroup_dict["space_group_number"] = 145
        elif "P41212" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "tetragonal"
            spacegroup_dict["point_group"]        = "4/mmm"
            spacegroup_dict["space_group"]        = "P41212"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = "unique_axis = c"
            spacegroup_dict["space_group_number"] = 92
        elif "P42212" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "tetragonal"
            spacegroup_dict["point_group"]        = "4/mmm"
            spacegroup_dict["space_group"]        = "P42212"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = "unique_axis = c"
            spacegroup_dict["space_group_number"] = 94
        elif "P43212" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "tetragonal"
            spacegroup_dict["point_group"]        = "4/mmm"
            spacegroup_dict["space_group"]        = "P43212"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = "unique_axis = c"
            spacegroup_dict["space_group_number"] = 96
        elif "P6522" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "hexagonal"
            spacegroup_dict["point_group"]        = "6/mmm"
            spacegroup_dict["space_group"]        = "P6522"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = "unique_axis = c"
            spacegroup_dict["space_group_number"] = 179
        elif "C2221" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "orthorhombic"
            spacegroup_dict["point_group"]        = "mmm"
            spacegroup_dict["space_group"]        = "C2221"
            spacegroup_dict["centering"]          = "C"
            spacegroup_dict["unique_axis"]        = ""
            spacegroup_dict["space_group_number"] = 20
        elif "C222" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "orthorhombic"
            spacegroup_dict["point_group"]        = "mmm"
            spacegroup_dict["space_group"]        = "C222"
            spacegroup_dict["centering"]          = "C"
            spacegroup_dict["unique_axis"]        = ""
            spacegroup_dict["space_group_number"] = 21
        elif "C2" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "monoclinic"
            spacegroup_dict["point_group"]        = "2/m_uab"
            spacegroup_dict["space_group"]        = "C2"
            spacegroup_dict["centering"]          = "C"
            ## spacegroup_dict["unique_axis"]     = ""
            spacegroup_dict["unique_axis"]        = "unique_axis = b"
            spacegroup_dict["space_group_number"] = 5
        elif "I222" in proc_params.space_group:
            spacegroup_dict["lattice_type"]       = "orthorhombic"
            spacegroup_dict["point_group"]        = "mmm"
            spacegroup_dict["space_group"]        = "I222"
            spacegroup_dict["centering"]          = "I"
            spacegroup_dict["unique_axis"]        = ""
            spacegroup_dict["space_group_number"] = 23
        else:
            spacegroup_dict["lattice_type"]       = "triclinic"
            spacegroup_dict["point_group"]        = "1"
            spacegroup_dict["space_group"]        = "P1"
            spacegroup_dict["centering"]          = "P"
            spacegroup_dict["unique_axis"]        = ""
            spacegroup_dict["space_group_number"] = 1

        return spacegroup_dict

    def batch_processed(self, batch):
        """Method called from EDNA via xmlrpc to set results

        :param batch: list of dictionaries describing processing results
        :type batch: lis
        """
        if self.started and (type(batch) in (tuple, list)):
            if type(batch[0]) not in (tuple, list):
                batch = [batch]

            self.batch_count = self.batch_count + len(batch)

            if len(batch) >= self.params_dict["images_num"]:
                self.all_frames_batch_processed = True
                logging.getLogger("HWR").info("All dozor scores %s of %s"%(len(batch), self.params_dict["images_num"]))
                self.finish_processing()
            else:
                logging.getLogger("HWR").info("Dozor scores %s of %s %s"%(len(batch), self.params_dict["images_num"],self.batch_count))

            for image in batch:
                frame_num = int(image[0])
                self.results_raw["spots_num"][frame_num] = image[1]
                if image[3] != 0:
                    self.results_raw["spots_resolution"][frame_num] = 1 / image[3]
                self.results_raw["score"][frame_num] = image[2]

                for score_key in self.results_raw.keys():
                    self.align_result(score_key, frame_num)

    def align_result(self, score_key, frame_num):
        if self.params_dict["lines_num"] > 1:
            col, row = self.grid.get_col_row_from_image(frame_num)
            self.results_aligned[score_key][col][row] = self.results_raw[score_key][frame_num]
        else:
            self.results_aligned[score_key][frame_num] = self.results_raw[score_key][frame_num]

    def dozor_is_changed(self, is_values):
        if self.started:
            for is_value in is_values:
                if isinstance(is_value,(list,tuple)):
                    self.results_raw["is"][is_value[0]] = is_value[1]
                    self.align_result("is", is_value[0])
                else:
                    self.results_raw["is"][is_values[0]] = is_values[1]
                    self.align_result("is", is_values[0])

            if isinstance(is_values,tuple):
                self.is_count = self.is_count + 1
            elif isinstance(is_values,list):
                self.is_count = self.is_count + len(is_values)

            if len(is_values) >= self.params_dict["images_num"]:
                self.all_frames_dozor_is = True
                logging.getLogger("HWR").info("All ISs %s of %s"%(len(is_values), self.params_dict["images_num"]))
                self.finish_processing()
            else:
                logging.getLogger("HWR").info("IS values %s of %s %s"%(len(is_values), self.params_dict["images_num"], self.is_count))


    def finish_processing(self):
        if self.all_frames_dozor_is and self.all_frames_batch_processed:
           logging.getLogger("HWR").info("Processing finished")
           self.set_processing_status("Success")
           self.all_frames_dozor_is = False
           self.all_frames_batch_processed = False
           self.is_count = 0
           self.batch_count = 0 

    def dozor_average_i_changed(self, average_i_value):
        if self.started:
            self.results_raw["average_intensity"] = np.append(
                self.results_raw["average_intensity"], average_i_value
            )
            self.results_aligned["average_intensity"] = np.append(
                self.results_aligned["average_intensity"], average_i_value
            )

    def update_map(self):
        return
        """
        Emits heat map update signal
        :return:
        """
        gevent.sleep(1)
        while self.started:
            self.emit("processingResultsUpdate", False)
            if self.params_dict["lines_num"] > 1:
                self.grid.set_score(self.results_raw["score"])
            gevent.sleep(0.5)

    def store_processing_results(self, status):
        """
        Stors processing results
        :param status: str
        :return:
        """
        GenericOnlineProcessing.store_processing_results(self, status)
        self.display_task.kill()
        gevent.spawn(self.create_hit_list_files)

    def store_result_xml(self):
        """
        Stores results in xml for further usage
        :return:
        """
        processing_xml_filename = os.path.join(
            self.params_dict["process_directory"], "dozor_result.xml"
        )
        dozor_result = XSDataResultControlDozor()
        for index in range(self.params_dict["images_num"]):
            dozor_image = XSDataControlImageDozor()
            dozor_image.setNumber(XSDataInteger(index))
            dozor_image.setScore(XSDataDouble(self.results_raw["score"][index]))
            dozor_image.setSpots_num_of(
                XSDataInteger(self.results_raw["spots_num"][index])
            )
            dozor_image.setSpots_resolution(
                XSDataDouble(self.results_raw["spots_resolution"][index])
            )
            dozor_result.addImageDozor(dozor_image)
        dozor_result.exportToFile(processing_xml_filename)
        logging.getLogger("HWR").info(
            "Online processing: Results saved in %s" % processing_xml_filename
        )

    def create_all_file_list(self):
        all_file_filename = os.path.join(
            self.params_dict["process_directory"], "images_fullpath.lst"
        )
        tpfirst_file_filename = os.path.join(
            self.params_dict["process_directory"], "images_fullpath_tpfirst.lst"
        )
        tplast_file_filename = os.path.join(
            self.params_dict["process_directory"], "images_fullpath_tplast.lst"
        )

        try:
            lst_file = open(all_file_filename, "w")
            for index in range(self.params_dict["images_num"]):
                lst_file.write(
                    self.params_dict["template"]
                    % (self.params_dict["run_number"], index + 1)
                    + "\n"
                 )
            self.print_log(
                "HWR",
                "debug",
                "Online processing: All image list stored in %s" % all_file_filename,
            )
        except BaseException:
            self.print_log(
                "GUI",
                "error",
                "Online processing: Unable to store all-image list in %s" % all_file_filename,
            )
        finally:
            lst_file.close()

        if self.params_dict["num_images_per_trigger"] > 3:
            try:
                lst_file_tpfirst = open(tpfirst_file_filename, "w")
                lst_file_tplast  = open(tplast_file_filename, "w")

                for index in range(self.params_dict["images_num"] / self.params_dict["num_images_per_trigger"]):
                    index_tpfirst = ( index      * self.params_dict["num_images_per_trigger"]) + 1
                    index_tplast  = ((index + 1) * self.params_dict["num_images_per_trigger"])


                    lst_file_tpfirst.write(
                        self.params_dict["template"]
                        % (self.params_dict["run_number"], index_tpfirst )
                        + "\n"
                    )
                    lst_file_tplast.write(
                        self.params_dict["template"]
                        % (self.params_dict["run_number"], index_tplast )
                        + "\n"
                    )

                self.print_log(
                    "HWR",
                    "debug",
                    "Online processing: First/last timepoint image lists stored in %s and %s" % (tpfirst_file_filename,tplast_file_filename),
                )
            except BaseException:
                self.print_log(
                    "GUI",
                    "error",
                    "Online processing: Unable to store image list of first/last timepoints.",
                )
            finally:
                lst_file_tpfirst.close()
                lst_file_tplast.close()

        return all_file_filename

    def create_hit_list_files(self):
        """
        :return:
        """
        lst_filename = os.path.join(
            self.params_dict["process_directory"], "images_dozor_hits_fullpath.lst"
        )
        nxds_filename = os.path.join(
            self.params_dict["process_directory"], "images_dozor_hits_nopath.lst"
        )

        dozor_results_filename = os.path.join(
            self.params_dict["process_directory"], "dozor_results.lst"
        )

        num_dozor_hits = 0

        try:
            lst_file = open(lst_filename, "w")
            nxds_lst_file = open(nxds_filename, "w")
            dozor_resolution_file = open(dozor_results_filename, "w")

            for index in range(self.params_dict["images_num"]):
                if self.results_raw["score"][index] > 0:
                    filename = self.params_dict["template"] % \
                        (self.params_dict["run_number"], index + 1)
                    lst_file.write(filename + "\n")
                    nxds_lst_file.write(os.path.basename(filename) + "\n")
                    # ADD number of spots and score here
                    dozor_resolution_file.write("%s %s %s %s\n" % (
                        filename,
                        self.results_raw["score"][index],
                        self.results_raw["spots_num"][index],
                        self.results_raw["spots_resolution"][index]
                        ))

                    num_dozor_hits = num_dozor_hits +1
            self.print_log(
                "GUI",
                "info",
                "Online processing: Hit list for crystfel stored in %s" % lst_filename,
            )
            self.print_log(
                "GUI",
                "info",
                "Online processing: Hit list for nxds stored in %s" % nxds_filename,
            )
            self.print_log(
                "GUI",
                "info",
                "Online processing: DOZOR results stored in %s" % dozor_results_filename,
            )


        except Exception:
            self.print_log(
                "GUI",
                "error",
                "Online processing: Unable to store hit list in %s" % lst_filename,
            )

        finally:
            lst_file.close()
            nxds_lst_file.close()
            dozor_resolution_file.close()

        msg = "Online processing: found %d dozor hits in %d collected images (%.2f%%)" % (
            num_dozor_hits,
            self.params_dict["images_num"],
            (100.0 * num_dozor_hits / self.params_dict["images_num"])
        )
        self.print_log("GUI", "info", msg)


    def start_crystfel_autoproc(self, all_file_filename):
        """
        Start crystfel processing
        :return:
        """
        acq_params = self.data_collection.acquisitions[0].acquisition_parameters
        proc_params_list = self.data_collection.processing_parameters_list

        self.print_log(
            "GUI",
            "error",
            "Number of target unit cells: %f" % (len(proc_params_list)),
        )

        # proc_params_list is a list of processing parameters

        if proc_params_list:
             # proc_params = proc_params_list[0]

             for proc_params in proc_params_list:
                 self.start_crystfel_autoproc_launchjob(proc_params, all_file_filename)
        else:
             proc_params = self.data_collection.processing_parameters
             self.start_crystfel_autoproc_launchjob(proc_params, all_file_filename)

    def start_crystfel_autoproc_launchjob(self, proc_params, all_file_filename):
        if proc_params.tag != "":
            cell_name_tag = "_" + proc_params.tag
        else:
            cell_name_tag = ""

        acq_params = self.data_collection.acquisitions[0].acquisition_parameters

        sample_basename   = self.params_dict["template"].split("/")[-1].split("_%d_%0")[0]
        # stream_filename   = sample_basename + "_crystfel_xgandalf.stream" 
        stream_filename   = sample_basename + cell_name_tag + "_crystfel.stream"
        geom_filename     = "crystfel_detector.geom"
        cell_filename     = "crystfel_cell" + cell_name_tag + ".cell"
        crystfel_autoproc = os.path.join(self.params_dict["process_directory"], "crystfel_autoproc" + cell_name_tag + ".sh" )
        log_filename      = "crystfel_xgandalf" + cell_name_tag + ".log"
        nxdsinp_filename  = os.path.join(self.params_dict["process_directory"], "nXDS.INP" )
        procdir_cluster   = self.params_dict["process_directory"].replace("/data/users/", "/home/")

        cell_filename     = cell_filename.replace("_cell_cell", "_cell")

        beam_x, beam_y = self.detector_hwobj.get_beam_centre()
        pixel_size_mm_x, pixel_size_mm_y = self.detector_hwobj.get_pixel_size_mm()

        self.print_log("HWR", "debug", "detector:          " + str(self.detector_hwobj.getProperty("type")) )
        self.print_log("HWR", "debug", "resolution cutoff: " + str(proc_params.resolution_cutoff) )
        self.print_log("HWR", "debug", "space group:       " + str(proc_params.space_group) )
        self.print_log("HWR", "debug", "PDB file:          " + str(proc_params.pdb_file) )
        self.print_log("HWR", "debug", "unit cell:         " + str(proc_params.cell_a) +", "+ str(proc_params.cell_b) +", "+ str(proc_params.cell_c) )

        ### Eiger 4M:
        detector_size_x = 2070
        detector_size_y = 2167
        ### Pilatus 2M:
        # detector_size_x = 1475
        # detector_size_y = 1679

        geom_file = CRYSTFEL_GEOM_FILE_TEMPLATE.format(
           self.detector_hwobj.get_distance() / 1000.,
           acq_params.energy * 1000,
           1000. / pixel_size_mm_x,
           detector_size_x -1,
           detector_size_y -1,
           -beam_x / pixel_size_mm_x,
           -beam_y / pixel_size_mm_y
        )

        data_file = open(os.path.join(self.params_dict["process_directory"], geom_filename), "w")
        data_file.write(geom_file)
        data_file.close()




        spacegroup_dict     = self.get_spacegroup_dict(proc_params, proc_params.space_group)

        lattice_type        = spacegroup_dict["lattice_type"]
        point_group         = spacegroup_dict["point_group"]
        space_group         = spacegroup_dict["space_group"]
        centering           = spacegroup_dict["centering"]
        unique_axis         = spacegroup_dict["unique_axis"]
        space_group_number  = spacegroup_dict["space_group_number"]


        cell_a              = proc_params.cell_a
        cell_b              = proc_params.cell_b
        cell_c              = proc_params.cell_c
        cell_alpha          = proc_params.cell_alpha
        cell_beta           = proc_params.cell_beta
        cell_gamma          = proc_params.cell_gamma

        cell_file = CRYSTFEL_CELL_FILE_TEMPLATE.format(lattice_type, centering, unique_axis, cell_a, cell_b, cell_c, cell_alpha, cell_beta, cell_gamma)

        data_file = open(os.path.join(self.params_dict["process_directory"], cell_filename), "w")
        data_file.write(cell_file)
        data_file.close()

        # print self.detector_hwobj.get_distance()
        wavelength=1e10 * CONST_H * CONST_C / (acq_params.energy * 1000)
        # print self.nxds_input_template

        nxds_file = self.nxds_input_template.format(
           image_template=self.params_dict["template"],
           image_directory=os.path.dirname(self.params_dict["template"]),
           image_list="nxds_dozor_hits.lst",
           space_group_number=int(space_group_number),
           cell_a=cell_a,
           cell_b=cell_b,
           cell_c=cell_c,
           cell_alpha=cell_alpha,
           cell_beta=cell_beta,
           cell_gamma=cell_gamma,
           wavelength=wavelength,
           pixel_size_mm_x=pixel_size_mm_x,
           pixel_size_mm_y=pixel_size_mm_y,
           detector_size_x=detector_size_x,
           detector_size_y=detector_size_y,
           org_x=beam_x / pixel_size_mm_x,
           org_y=beam_y / pixel_size_mm_y,
           detector_distance=self.detector_hwobj.get_distance()
        )

        data_file = open(nxdsinp_filename, "w")
        data_file.write(nxds_file)
        data_file.close()

        end_of_line_to_execute = " %s %s %s %s %s %s %s %.2f" % (
            self.params_dict["process_directory"],
            all_file_filename,
            geom_filename,
            stream_filename,
            cell_filename,
            self.crystfel_params["point_group"],
            proc_params.space_group,
            acq_params.resolution,
        )
        self.print_log("HWR", "debug", "Online processing: Starting crystfel %s with parameters %s " % (self.crystfel_script, end_of_line_to_execute) )

        # num_cores_crystfel    = 160
        # num_cores_partialator =  50

        jobname = stream_filename.split("_crystfel_")[0]
        self.print_log( "GUI", "error", "Online processing jobname: %s " % (jobname) )
        jobname = "CF_" + sample_basename + cell_name_tag
        self.print_log( "GUI", "error", "Online processing jobname: %s " % (jobname) )

        # crystfel_command = "ssh bcrunch 'cd %s; bash crystfel_autoproc.sh'" % procdir_cluster
        # crystfel_command = "ssh bcrunch 'sbatch -c %d -t 3-00:00 -D %s -J %s %s/crystfel_autoproc.sh'" % (self.crystfel_params["num_cores_crystfel"], procdir_cluster, jobname, procdir_cluster)
        crystfel_command = "ssh bcrunch 'sbatch -c %d -t 3-00:00 -D %s -J %s %s'" % (self.crystfel_params["num_cores_crystfel"], procdir_cluster, jobname, procdir_cluster + "/crystfel_autoproc" + cell_name_tag + ".sh")

        if self.params_dict["num_images_per_trigger"] > 3:
            self.print_log( "HWR", "debug", "Trying to launch first/last timepoint jobs..." )
            # crystfel_command_tp_first = "ssh bcrunch 'sbatch -c %d -t 3-00:00 -D %s -J %s crystfel_autoproc.sh'" % (self.crystfel_params["num_cores_crystfel"] /2, procdir_cluster, jobname + "_fst")
            # crystfel_command_tp_last  = "ssh bcrunch 'sbatch -c %d -t 3-00:00 -D %s -J %s crystfel_autoproc.sh'" % (self.crystfel_params["num_cores_crystfel"] /2, procdir_cluster, jobname + "_lst")

        self.print_log("HWR", "debug", "HARE_num, num_images_per_trigger: %f. %f" % (self.params_dict["hare_num"], self.params_dict["num_images_per_trigger"]) )

        crystfel_script = self.crystfel_script_template.format(
            num_cores_crystfel=self.crystfel_params["num_cores_crystfel"],
            procdir_cluster=procdir_cluster,
            geom_filename=geom_filename,
            stream_filename=stream_filename,
            cell_filename=cell_filename,
            log_filename=log_filename,
            num_cores_partialator=self.crystfel_params["num_cores_partialator"],
            pdb_file=proc_params.pdb_file,
            pdb_file_end=str(proc_params.pdb_file).split("/")[-1],
            resolution="{:.1f}".format(proc_params.resolution_cutoff),
            hare_number=self.params_dict["hare_num"],
            burst_number=self.params_dict["num_images_per_trigger"],
            ## burst_number=1,
            pointgroup=point_group,
            spacegroup=space_group
        )

        data_file = open(crystfel_autoproc, "w")
        data_file.write(crystfel_script)
        data_file.close()

        self.print_log("HWR", "debug", "Online processing: crystfel command: \n %s" % (crystfel_command) )

        subprocess.Popen(crystfel_command, shell=True, stdin=None, stdout=None, stderr=None, close_fds=True)




#!/usr/bin/env python

#
# Generated Fri Feb 20 04:42::27 2015 by EDGenerateDS.
#

import os
import sys
from xml.dom import (
    Node,
    minidom,
)

from XSDataCommon import (
    XSDataBoolean,
    XSDataDouble,
    XSDataFile,
    XSDataInput,
    XSDataInteger,
    XSDataResult,
    XSDataString,
)

strEdnaHome = os.environ.get("EDNA_HOME", None)

dictLocation = {
    "XSDataCommon": "kernel/datamodel/.",
    "XSDataCommon": "kernel/datamodel/.",
    "XSDataCommon": "kernel/datamodel/.",
    "XSDataCommon": "kernel/datamodel/.",
    "XSDataCommon": "kernel/datamodel/.",
    "XSDataCommon": "kernel/datamodel/.",
    "XSDataCommon": "kernel/datamodel/.",
}

try:
    from XSDataCommon import (
        XSDataBoolean,
        XSDataDouble,
        XSDataFile,
        XSDataInput,
        XSDataInteger,
        XSDataResult,
        XSDataString,
    )
except ImportError as error:
    if strEdnaHome is not None:
        for strXsdName in dictLocation:
            strXsdModule = strXsdName + ".py"
            strRootdir = os.path.dirname(
                os.path.abspath(os.path.join(strEdnaHome, dictLocation[strXsdName]))
            )
            for strRoot, listDirs, listFiles in os.walk(strRootdir):
                if strXsdModule in listFiles:
                    sys.path.append(strRoot)
    else:
        raise error


#
# Support/utility functions.
#

# Compabiltity between Python 2 and 3:
if sys.version.startswith("3"):
    unicode = str
    from io import StringIO
else:
    from StringIO import StringIO


def showIndent(outfile, level):
    for idx in range(level):
        outfile.write(unicode("    "))


def warnEmptyAttribute(_strName, _strTypeName):
    pass
    # if not _strTypeName in ["float", "double", "string", "boolean", "integer"]:
    #    print("Warning! Non-optional attribute %s of type %s is None!" % (_strName, _strTypeName))


class MixedContainer(object):
    # Constants for category:
    CategoryNone = 0
    CategoryText = 1
    CategorySimple = 2
    CategoryComplex = 3
    # Constants for content_type:
    TypeNone = 0
    TypeText = 1
    TypeString = 2
    TypeInteger = 3
    TypeFloat = 4
    TypeDecimal = 5
    TypeDouble = 6
    TypeBoolean = 7

    def __init__(self, category, content_type, name, value):
        self.category = category
        self.content_type = content_type
        self.name = name
        self.value = value

    def getCategory(self):
        return self.category

    def getContenttype(self, content_type):
        return self.content_type

    def getValue(self):
        return self.value

    def getName(self):
        return self.name

    def export(self, outfile, level, name):
        if self.category == MixedContainer.CategoryText:
            outfile.write(self.value)
        elif self.category == MixedContainer.CategorySimple:
            self.exportSimple(outfile, level, name)
        else:  # category == MixedContainer.CategoryComplex
            self.value.export(outfile, level, name)

    def exportSimple(self, outfile, level, name):
        if self.content_type == MixedContainer.TypeString:
            outfile.write(unicode("<%s>%s</%s>" % (self.name, self.value, self.name)))
        elif (
            self.content_type == MixedContainer.TypeInteger
            or self.content_type == MixedContainer.TypeBoolean
        ):
            outfile.write(unicode("<%s>%d</%s>" % (self.name, self.value, self.name)))
        elif (
            self.content_type == MixedContainer.TypeFloat
            or self.content_type == MixedContainer.TypeDecimal
        ):
            outfile.write(unicode("<%s>%f</%s>" % (self.name, self.value, self.name)))
        elif self.content_type == MixedContainer.TypeDouble:
            outfile.write(unicode("<%s>%g</%s>" % (self.name, self.value, self.name)))


#
# Data representation classes.
#


class XSDataControlImageDozor(object):
    def __init__(
        self,
        score=None,
        powder_wilson_rfactor=None,
        powder_wilson_correlation=None,
        powder_wilson_resolution=None,
        powder_wilson_bfactor=None,
        powder_wilson_scale=None,
        spots_resolution=None,
        spots_int_aver=None,
        spots_num_of=None,
        number=None,
        image=None,
    ):
        if image is None:
            self._image = None
        elif image.__class__.__name__ == "XSDataFile":
            self._image = image
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor constructor argument 'image' is not XSDataFile but %s"
                % self._image.__class__.__name__
            )
            raise Exception(strMessage)
        if number is None:
            self._number = None
        elif number.__class__.__name__ == "XSDataInteger":
            self._number = number
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor constructor argument 'number' is not XSDataInteger but %s"
                % self._number.__class__.__name__
            )
            raise Exception(strMessage)
        if spots_num_of is None:
            self._spots_num_of = None
        elif spots_num_of.__class__.__name__ == "XSDataInteger":
            self._spots_num_of = spots_num_of
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor constructor argument 'spots_num_of' is not XSDataInteger but %s"
                % self._spots_num_of.__class__.__name__
            )
            raise Exception(strMessage)
        if spots_int_aver is None:
            self._spots_int_aver = None
        elif spots_int_aver.__class__.__name__ == "XSDataDouble":
            self._spots_int_aver = spots_int_aver
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor constructor argument 'spots_int_aver' is not XSDataDouble but %s"
                % self._spots_int_aver.__class__.__name__
            )
            raise Exception(strMessage)
        if spots_resolution is None:
            self._spots_resolution = None
        elif spots_resolution.__class__.__name__ == "XSDataDouble":
            self._spots_resolution = spots_resolution
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor constructor argument 'spots_resolution' is not XSDataDouble but %s"
                % self._spots_resolution.__class__.__name__
            )
            raise Exception(strMessage)
        if powder_wilson_scale is None:
            self._powder_wilson_scale = None
        elif powder_wilson_scale.__class__.__name__ == "XSDataDouble":
            self._powder_wilson_scale = powder_wilson_scale
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor constructor argument 'powder_wilson_scale' is not XSDataDouble but %s"
                % self._powder_wilson_scale.__class__.__name__
            )
            raise Exception(strMessage)
        if powder_wilson_bfactor is None:
            self._powder_wilson_bfactor = None
        elif powder_wilson_bfactor.__class__.__name__ == "XSDataDouble":
            self._powder_wilson_bfactor = powder_wilson_bfactor
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor constructor argument 'powder_wilson_bfactor' is not XSDataDouble but %s"
                % self._powder_wilson_bfactor.__class__.__name__
            )
            raise Exception(strMessage)
        if powder_wilson_resolution is None:
            self._powder_wilson_resolution = None
        elif powder_wilson_resolution.__class__.__name__ == "XSDataDouble":
            self._powder_wilson_resolution = powder_wilson_resolution
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor constructor argument 'powder_wilson_resolution' is not XSDataDouble but %s"
                % self._powder_wilson_resolution.__class__.__name__
            )
            raise Exception(strMessage)
        if powder_wilson_correlation is None:
            self._powder_wilson_correlation = None
        elif powder_wilson_correlation.__class__.__name__ == "XSDataDouble":
            self._powder_wilson_correlation = powder_wilson_correlation
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor constructor argument 'powder_wilson_correlation' is not XSDataDouble but %s"
                % self._powder_wilson_correlation.__class__.__name__
            )
            raise Exception(strMessage)
        if powder_wilson_rfactor is None:
            self._powder_wilson_rfactor = None
        elif powder_wilson_rfactor.__class__.__name__ == "XSDataDouble":
            self._powder_wilson_rfactor = powder_wilson_rfactor
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor constructor argument 'powder_wilson_rfactor' is not XSDataDouble but %s"
                % self._powder_wilson_rfactor.__class__.__name__
            )
            raise Exception(strMessage)
        if score is None:
            self._score = None
        elif score.__class__.__name__ == "XSDataDouble":
            self._score = score
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor constructor argument 'score' is not XSDataDouble but %s"
                % self._score.__class__.__name__
            )
            raise Exception(strMessage)

    # Methods and properties for the 'image' attribute
    def getImage(self):
        return self._image

    def setImage(self, image):
        if image is None:
            self._image = None
        elif image.__class__.__name__ == "XSDataFile":
            self._image = image
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor.setImage argument is not XSDataFile but %s"
                % image.__class__.__name__
            )
            raise Exception(strMessage)

    def delImage(self):
        self._image = None

    image = property(getImage, setImage, delImage, "Property for image")
    # Methods and properties for the 'number' attribute

    def getNumber(self):
        return self._number

    def setNumber(self, number):
        if number is None:
            self._number = None
        elif number.__class__.__name__ == "XSDataInteger":
            self._number = number
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor.setNumber argument is not XSDataInteger but %s"
                % number.__class__.__name__
            )
            raise Exception(strMessage)

    def delNumber(self):
        self._number = None

    number = property(getNumber, setNumber, delNumber, "Property for number")
    # Methods and properties for the 'spots_num_of' attribute

    def getSpots_num_of(self):
        return self._spots_num_of

    def setSpots_num_of(self, spots_num_of):
        if spots_num_of is None:
            self._spots_num_of = None
        elif spots_num_of.__class__.__name__ == "XSDataInteger":
            self._spots_num_of = spots_num_of
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor.setSpots_num_of argument is not XSDataInteger but %s"
                % spots_num_of.__class__.__name__
            )
            raise Exception(strMessage)

    def delSpots_num_of(self):
        self._spots_num_of = None

    spots_num_of = property(
        getSpots_num_of, setSpots_num_of, delSpots_num_of, "Property for spots_num_of"
    )
    # Methods and properties for the 'spots_int_aver' attribute

    def getSpots_int_aver(self):
        return self._spots_int_aver

    def setSpots_int_aver(self, spots_int_aver):
        if spots_int_aver is None:
            self._spots_int_aver = None
        elif spots_int_aver.__class__.__name__ == "XSDataDouble":
            self._spots_int_aver = spots_int_aver
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor.setSpots_int_aver argument is not XSDataDouble but %s"
                % spots_int_aver.__class__.__name__
            )
            raise Exception(strMessage)

    def delSpots_int_aver(self):
        self._spots_int_aver = None

    spots_int_aver = property(
        getSpots_int_aver,
        setSpots_int_aver,
        delSpots_int_aver,
        "Property for spots_int_aver",
    )
    # Methods and properties for the 'spots_resolution' attribute

    def getSpots_resolution(self):
        return self._spots_resolution

    def setSpots_resolution(self, spots_resolution):
        if spots_resolution is None:
            self._spots_resolution = None
        elif spots_resolution.__class__.__name__ == "XSDataDouble":
            self._spots_resolution = spots_resolution
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor.setSpots_resolution argument is not XSDataDouble but %s"
                % spots_resolution.__class__.__name__
            )
            raise Exception(strMessage)

    def delSpots_resolution(self):
        self._spots_resolution = None

    spots_resolution = property(
        getSpots_resolution,
        setSpots_resolution,
        delSpots_resolution,
        "Property for spots_resolution",
    )
    # Methods and properties for the 'powder_wilson_scale' attribute

    def getPowder_wilson_scale(self):
        return self._powder_wilson_scale

    def setPowder_wilson_scale(self, powder_wilson_scale):
        if powder_wilson_scale is None:
            self._powder_wilson_scale = None
        elif powder_wilson_scale.__class__.__name__ == "XSDataDouble":
            self._powder_wilson_scale = powder_wilson_scale
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor.setPowder_wilson_scale argument is not XSDataDouble but %s"
                % powder_wilson_scale.__class__.__name__
            )
            raise Exception(strMessage)

    def delPowder_wilson_scale(self):
        self._powder_wilson_scale = None

    powder_wilson_scale = property(
        getPowder_wilson_scale,
        setPowder_wilson_scale,
        delPowder_wilson_scale,
        "Property for powder_wilson_scale",
    )
    # Methods and properties for the 'powder_wilson_bfactor' attribute

    def getPowder_wilson_bfactor(self):
        return self._powder_wilson_bfactor

    def setPowder_wilson_bfactor(self, powder_wilson_bfactor):
        if powder_wilson_bfactor is None:
            self._powder_wilson_bfactor = None
        elif powder_wilson_bfactor.__class__.__name__ == "XSDataDouble":
            self._powder_wilson_bfactor = powder_wilson_bfactor
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor.setPowder_wilson_bfactor argument is not XSDataDouble but %s"
                % powder_wilson_bfactor.__class__.__name__
            )
            raise Exception(strMessage)

    def delPowder_wilson_bfactor(self):
        self._powder_wilson_bfactor = None

    powder_wilson_bfactor = property(
        getPowder_wilson_bfactor,
        setPowder_wilson_bfactor,
        delPowder_wilson_bfactor,
        "Property for powder_wilson_bfactor",
    )
    # Methods and properties for the 'powder_wilson_resolution' attribute

    def getPowder_wilson_resolution(self):
        return self._powder_wilson_resolution

    def setPowder_wilson_resolution(self, powder_wilson_resolution):
        if powder_wilson_resolution is None:
            self._powder_wilson_resolution = None
        elif powder_wilson_resolution.__class__.__name__ == "XSDataDouble":
            self._powder_wilson_resolution = powder_wilson_resolution
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor.setPowder_wilson_resolution argument is not XSDataDouble but %s"
                % powder_wilson_resolution.__class__.__name__
            )
            raise Exception(strMessage)

    def delPowder_wilson_resolution(self):
        self._powder_wilson_resolution = None

    powder_wilson_resolution = property(
        getPowder_wilson_resolution,
        setPowder_wilson_resolution,
        delPowder_wilson_resolution,
        "Property for powder_wilson_resolution",
    )
    # Methods and properties for the 'powder_wilson_correlation' attribute

    def getPowder_wilson_correlation(self):
        return self._powder_wilson_correlation

    def setPowder_wilson_correlation(self, powder_wilson_correlation):
        if powder_wilson_correlation is None:
            self._powder_wilson_correlation = None
        elif powder_wilson_correlation.__class__.__name__ == "XSDataDouble":
            self._powder_wilson_correlation = powder_wilson_correlation
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor.setPowder_wilson_correlation argument is not XSDataDouble but %s"
                % powder_wilson_correlation.__class__.__name__
            )
            raise Exception(strMessage)

    def delPowder_wilson_correlation(self):
        self._powder_wilson_correlation = None

    powder_wilson_correlation = property(
        getPowder_wilson_correlation,
        setPowder_wilson_correlation,
        delPowder_wilson_correlation,
        "Property for powder_wilson_correlation",
    )
    # Methods and properties for the 'powder_wilson_rfactor' attribute

    def getPowder_wilson_rfactor(self):
        return self._powder_wilson_rfactor

    def setPowder_wilson_rfactor(self, powder_wilson_rfactor):
        if powder_wilson_rfactor is None:
            self._powder_wilson_rfactor = None
        elif powder_wilson_rfactor.__class__.__name__ == "XSDataDouble":
            self._powder_wilson_rfactor = powder_wilson_rfactor
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor.setPowder_wilson_rfactor argument is not XSDataDouble but %s"
                % powder_wilson_rfactor.__class__.__name__
            )
            raise Exception(strMessage)

    def delPowder_wilson_rfactor(self):
        self._powder_wilson_rfactor = None

    powder_wilson_rfactor = property(
        getPowder_wilson_rfactor,
        setPowder_wilson_rfactor,
        delPowder_wilson_rfactor,
        "Property for powder_wilson_rfactor",
    )
    # Methods and properties for the 'score' attribute

    def getScore(self):
        return self._score

    def setScore(self, score):
        if score is None:
            self._score = None
        elif score.__class__.__name__ == "XSDataDouble":
            self._score = score
        else:
            strMessage = (
                "ERROR! XSDataControlImageDozor.setScore argument is not XSDataDouble but %s"
                % score.__class__.__name__
            )
            raise Exception(strMessage)

    def delScore(self):
        self._score = None

    score = property(getScore, setScore, delScore, "Property for score")

    def export(self, outfile, level, name_="XSDataControlImageDozor"):
        showIndent(outfile, level)
        outfile.write(unicode("<%s>\n" % name_))
        self.exportChildren(outfile, level + 1, name_)
        showIndent(outfile, level)
        outfile.write(unicode("</%s>\n" % name_))

    def exportChildren(self, outfile, level, name_="XSDataControlImageDozor"):
        if self._image is not None:
            self.image.export(outfile, level, name_="image")
        else:
            warnEmptyAttribute("image", "XSDataFile")
        if self._number is not None:
            self.number.export(outfile, level, name_="number")
        if self._spots_num_of is not None:
            self.spots_num_of.export(outfile, level, name_="spots_num_of")
        else:
            warnEmptyAttribute("spots_num_of", "XSDataInteger")
        if self._spots_int_aver is not None:
            self.spots_int_aver.export(outfile, level, name_="spots_int_aver")
        else:
            warnEmptyAttribute("spots_int_aver", "XSDataDouble")
        if self._spots_resolution is not None:
            self.spots_resolution.export(outfile, level, name_="spots_resolution")
        if self._powder_wilson_scale is not None:
            self.powder_wilson_scale.export(outfile, level, name_="powder_wilson_scale")
        if self._powder_wilson_bfactor is not None:
            self.powder_wilson_bfactor.export(
                outfile, level, name_="powder_wilson_bfactor"
            )
        if self._powder_wilson_resolution is not None:
            self.powder_wilson_resolution.export(
                outfile, level, name_="powder_wilson_resolution"
            )
        if self._powder_wilson_correlation is not None:
            self.powder_wilson_correlation.export(
                outfile, level, name_="powder_wilson_correlation"
            )
        if self._powder_wilson_rfactor is not None:
            self.powder_wilson_rfactor.export(
                outfile, level, name_="powder_wilson_rfactor"
            )
        if self._score is not None:
            self.score.export(outfile, level, name_="score")

    def build(self, node_):
        for child_ in node_.childNodes:
            nodeName_ = child_.nodeName.split(":")[-1]
            self.buildChildren(child_, nodeName_)

    def buildChildren(self, child_, nodeName_):
        if child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "image":
            obj_ = XSDataFile()
            obj_.build(child_)
            self.setImage(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "number":
            obj_ = XSDataInteger()
            obj_.build(child_)
            self.setNumber(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "spots_num_of":
            obj_ = XSDataInteger()
            obj_.build(child_)
            self.setSpots_num_of(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "spots_int_aver":
            obj_ = XSDataDouble()
            obj_.build(child_)
            self.setSpots_int_aver(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "spots_resolution":
            obj_ = XSDataDouble()
            obj_.build(child_)
            self.setSpots_resolution(obj_)
        elif (
            child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "powder_wilson_scale"
        ):
            obj_ = XSDataDouble()
            obj_.build(child_)
            self.setPowder_wilson_scale(obj_)
        elif (
            child_.nodeType == Node.ELEMENT_NODE
            and nodeName_ == "powder_wilson_bfactor"
        ):
            obj_ = XSDataDouble()
            obj_.build(child_)
            self.setPowder_wilson_bfactor(obj_)
        elif (
            child_.nodeType == Node.ELEMENT_NODE
            and nodeName_ == "powder_wilson_resolution"
        ):
            obj_ = XSDataDouble()
            obj_.build(child_)
            self.setPowder_wilson_resolution(obj_)
        elif (
            child_.nodeType == Node.ELEMENT_NODE
            and nodeName_ == "powder_wilson_correlation"
        ):
            obj_ = XSDataDouble()
            obj_.build(child_)
            self.setPowder_wilson_correlation(obj_)
        elif (
            child_.nodeType == Node.ELEMENT_NODE
            and nodeName_ == "powder_wilson_rfactor"
        ):
            obj_ = XSDataDouble()
            obj_.build(child_)
            self.setPowder_wilson_rfactor(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "score":
            obj_ = XSDataDouble()
            obj_.build(child_)
            self.setScore(obj_)

    # Method for marshalling an object
    def marshal(self):
        oStreamString = StringIO()
        oStreamString.write(unicode('<?xml version="1.0" ?>\n'))
        self.export(oStreamString, 0, name_="XSDataControlImageDozor")
        oStringXML = oStreamString.getValue()
        oStreamString.close()
        return oStringXML

    # Only to export the entire XML tree to a file stream on disk
    def exportToFile(self, _outfileName):
        outfile = open(_outfileName, "w")
        outfile.write(unicode('<?xml version="1.0" ?>\n'))
        self.export(outfile, 0, name_="XSDataControlImageDozor")
        outfile.close()

    # Deprecated method, replaced by exportToFile
    def outputFile(self, _outfileName):
        print(
            "WARNING: Method outputFile in class XSDataControlImageDozor is deprecated, please use instead exportToFile!"
        )
        self.exportToFile(_outfileName)

    # Method for making a copy in a new instance
    def copy(self):
        return XSDataControlImageDozor.parseString(self.marshal())

    # Static method for parsing a string
    def parseString(_inString):
        doc = minidom.parseString(_inString)
        rootNode = doc.documentElement
        rootObj = XSDataControlImageDozor()
        rootObj.build(rootNode)
        # Check that all minOccurs are obeyed by marshalling the created object
        oStreamString = StringIO()
        rootObj.export(oStreamString, 0, name_="XSDataControlImageDozor")
        oStreamString.close()
        return rootObj

    parseString = staticmethod(parseString)
    # Static method for parsing a file

    def parseFile(_inFilePath):
        doc = minidom.parse(_inFilePath)
        rootNode = doc.documentElement
        rootObj = XSDataControlImageDozor()
        rootObj.build(rootNode)
        return rootObj

    parseFile = staticmethod(parseFile)


# end class XSDataControlImageDozor


class XSDataInputControlDozor(XSDataInput):
    def __init__(
        self,
        configuration=None,
        beamstopDistance=None,
        beamstopSize=None,
        beamstopDirection=None,
        pixelMax=None,
        pixelMin=None,
        reversing_rotation=None,
        line_number_of=None,
        last_run_number=None,
        first_run_number=None,
        last_image_number=None,
        first_image_number=None,
        template=None,
    ):
        XSDataInput.__init__(self, configuration)
        if template is None:
            self._template = None
        elif template.__class__.__name__ == "XSDataString":
            self._template = template
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor constructor argument 'template' is not XSDataString but %s"
                % self._template.__class__.__name__
            )
            raise Exception(strMessage)
        if first_image_number is None:
            self._first_image_number = None
        elif first_image_number.__class__.__name__ == "XSDataInteger":
            self._first_image_number = first_image_number
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor constructor argument 'first_image_number' is not XSDataInteger but %s"
                % self._first_image_number.__class__.__name__
            )
            raise Exception(strMessage)
        if last_image_number is None:
            self._last_image_number = None
        elif last_image_number.__class__.__name__ == "XSDataInteger":
            self._last_image_number = last_image_number
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor constructor argument 'last_image_number' is not XSDataInteger but %s"
                % self._last_image_number.__class__.__name__
            )
            raise Exception(strMessage)
        if first_run_number is None:
            self._first_run_number = None
        elif first_run_number.__class__.__name__ == "XSDataInteger":
            self._first_run_number = first_run_number
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor constructor argument 'first_run_number' is not XSDataInteger but %s"
                % self._first_run_number.__class__.__name__
            )
            raise Exception(strMessage)
        if last_run_number is None:
            self._last_run_number = None
        elif last_run_number.__class__.__name__ == "XSDataInteger":
            self._last_run_number = last_run_number
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor constructor argument 'last_run_number' is not XSDataInteger but %s"
                % self._last_run_number.__class__.__name__
            )
            raise Exception(strMessage)
        if line_number_of is None:
            self._line_number_of = None
        elif line_number_of.__class__.__name__ == "XSDataInteger":
            self._line_number_of = line_number_of
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor constructor argument 'line_number_of' is not XSDataInteger but %s"
                % self._line_number_of.__class__.__name__
            )
            raise Exception(strMessage)
        if reversing_rotation is None:
            self._reversing_rotation = None
        elif reversing_rotation.__class__.__name__ == "XSDataBoolean":
            self._reversing_rotation = reversing_rotation
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor constructor argument 'reversing_rotation' is not XSDataBoolean but %s"
                % self._reversing_rotation.__class__.__name__
            )
            raise Exception(strMessage)
        if pixelMin is None:
            self._pixelMin = None
        elif pixelMin.__class__.__name__ == "XSDataInteger":
            self._pixelMin = pixelMin
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor constructor argument 'pixelMin' is not XSDataInteger but %s"
                % self._pixelMin.__class__.__name__
            )
            raise Exception(strMessage)
        if pixelMax is None:
            self._pixelMax = None
        elif pixelMax.__class__.__name__ == "XSDataInteger":
            self._pixelMax = pixelMax
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor constructor argument 'pixelMax' is not XSDataInteger but %s"
                % self._pixelMax.__class__.__name__
            )
            raise Exception(strMessage)
        if beamstopDirection is None:
            self._beamstopDirection = None
        elif beamstopDirection.__class__.__name__ == "XSDataString":
            self._beamstopDirection = beamstopDirection
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor constructor argument 'beamstopDirection' is not XSDataString but %s"
                % self._beamstopDirection.__class__.__name__
            )
            raise Exception(strMessage)
        if beamstopSize is None:
            self._beamstopSize = None
        elif beamstopSize.__class__.__name__ == "XSDataDouble":
            self._beamstopSize = beamstopSize
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor constructor argument 'beamstopSize' is not XSDataDouble but %s"
                % self._beamstopSize.__class__.__name__
            )
            raise Exception(strMessage)
        if beamstopDistance is None:
            self._beamstopDistance = None
        elif beamstopDistance.__class__.__name__ == "XSDataDouble":
            self._beamstopDistance = beamstopDistance
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor constructor argument 'beamstopDistance' is not XSDataDouble but %s"
                % self._beamstopDistance.__class__.__name__
            )
            raise Exception(strMessage)

    # Methods and properties for the 'template' attribute
    def getTemplate(self):
        return self._template

    def setTemplate(self, template):
        if template is None:
            self._template = None
        elif template.__class__.__name__ == "XSDataString":
            self._template = template
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor.setTemplate argument is not XSDataString but %s"
                % template.__class__.__name__
            )
            raise Exception(strMessage)

    def delTemplate(self):
        self._template = None

    template = property(getTemplate, setTemplate, delTemplate, "Property for template")
    # Methods and properties for the 'first_image_number' attribute

    def getFirst_image_number(self):
        return self._first_image_number

    def setFirst_image_number(self, first_image_number):
        if first_image_number is None:
            self._first_image_number = None
        elif first_image_number.__class__.__name__ == "XSDataInteger":
            self._first_image_number = first_image_number
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor.setFirst_image_number argument is not XSDataInteger but %s"
                % first_image_number.__class__.__name__
            )
            raise Exception(strMessage)

    def delFirst_image_number(self):
        self._first_image_number = None

    first_image_number = property(
        getFirst_image_number,
        setFirst_image_number,
        delFirst_image_number,
        "Property for first_image_number",
    )
    # Methods and properties for the 'last_image_number' attribute

    def getLast_image_number(self):
        return self._last_image_number

    def setLast_image_number(self, last_image_number):
        if last_image_number is None:
            self._last_image_number = None
        elif last_image_number.__class__.__name__ == "XSDataInteger":
            self._last_image_number = last_image_number
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor.setLast_image_number argument is not XSDataInteger but %s"
                % last_image_number.__class__.__name__
            )
            raise Exception(strMessage)

    def delLast_image_number(self):
        self._last_image_number = None

    last_image_number = property(
        getLast_image_number,
        setLast_image_number,
        delLast_image_number,
        "Property for last_image_number",
    )
    # Methods and properties for the 'first_run_number' attribute

    def getFirst_run_number(self):
        return self._first_run_number

    def setFirst_run_number(self, first_run_number):
        if first_run_number is None:
            self._first_run_number = None
        elif first_run_number.__class__.__name__ == "XSDataInteger":
            self._first_run_number = first_run_number
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor.setFirst_run_number argument is not XSDataInteger but %s"
                % first_run_number.__class__.__name__
            )
            raise Exception(strMessage)

    def delFirst_run_number(self):
        self._first_run_number = None

    first_run_number = property(
        getFirst_run_number,
        setFirst_run_number,
        delFirst_run_number,
        "Property for first_run_number",
    )
    # Methods and properties for the 'last_run_number' attribute

    def getLast_run_number(self):
        return self._last_run_number

    def setLast_run_number(self, last_run_number):
        if last_run_number is None:
            self._last_run_number = None
        elif last_run_number.__class__.__name__ == "XSDataInteger":
            self._last_run_number = last_run_number
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor.setLast_run_number argument is not XSDataInteger but %s"
                % last_run_number.__class__.__name__
            )
            raise Exception(strMessage)

    def delLast_run_number(self):
        self._last_run_number = None

    last_run_number = property(
        getLast_run_number,
        setLast_run_number,
        delLast_run_number,
        "Property for last_run_number",
    )
    # Methods and properties for the 'line_number_of' attribute

    def getLine_number_of(self):
        return self._line_number_of

    def setLine_number_of(self, line_number_of):
        if line_number_of is None:
            self._line_number_of = None
        elif line_number_of.__class__.__name__ == "XSDataInteger":
            self._line_number_of = line_number_of
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor.setLine_number_of argument is not XSDataInteger but %s"
                % line_number_of.__class__.__name__
            )
            raise Exception(strMessage)

    def delLine_number_of(self):
        self._line_number_of = None

    line_number_of = property(
        getLine_number_of,
        setLine_number_of,
        delLine_number_of,
        "Property for line_number_of",
    )
    # Methods and properties for the 'reversing_rotation' attribute

    def getReversing_rotation(self):
        return self._reversing_rotation

    def setReversing_rotation(self, reversing_rotation):
        if reversing_rotation is None:
            self._reversing_rotation = None
        elif reversing_rotation.__class__.__name__ == "XSDataBoolean":
            self._reversing_rotation = reversing_rotation
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor.setReversing_rotation argument is not XSDataBoolean but %s"
                % reversing_rotation.__class__.__name__
            )
            raise Exception(strMessage)

    def delReversing_rotation(self):
        self._reversing_rotation = None

    reversing_rotation = property(
        getReversing_rotation,
        setReversing_rotation,
        delReversing_rotation,
        "Property for reversing_rotation",
    )
    # Methods and properties for the 'pixelMin' attribute

    def getPixelMin(self):
        return self._pixelMin

    def setPixelMin(self, pixelMin):
        if pixelMin is None:
            self._pixelMin = None
        elif pixelMin.__class__.__name__ == "XSDataInteger":
            self._pixelMin = pixelMin
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor.setPixelMin argument is not XSDataInteger but %s"
                % pixelMin.__class__.__name__
            )
            raise Exception(strMessage)

    def delPixelMin(self):
        self._pixelMin = None

    pixelMin = property(getPixelMin, setPixelMin, delPixelMin, "Property for pixelMin")
    # Methods and properties for the 'pixelMax' attribute

    def getPixelMax(self):
        return self._pixelMax

    def setPixelMax(self, pixelMax):
        if pixelMax is None:
            self._pixelMax = None
        elif pixelMax.__class__.__name__ == "XSDataInteger":
            self._pixelMax = pixelMax
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor.setPixelMax argument is not XSDataInteger but %s"
                % pixelMax.__class__.__name__
            )
            raise Exception(strMessage)

    def delPixelMax(self):
        self._pixelMax = None

    pixelMax = property(getPixelMax, setPixelMax, delPixelMax, "Property for pixelMax")
    # Methods and properties for the 'beamstopDirection' attribute

    def getBeamstopDirection(self):
        return self._beamstopDirection

    def setBeamstopDirection(self, beamstopDirection):
        if beamstopDirection is None:
            self._beamstopDirection = None
        elif beamstopDirection.__class__.__name__ == "XSDataString":
            self._beamstopDirection = beamstopDirection
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor.setBeamstopDirection argument is not XSDataString but %s"
                % beamstopDirection.__class__.__name__
            )
            raise Exception(strMessage)

    def delBeamstopDirection(self):
        self._beamstopDirection = None

    beamstopDirection = property(
        getBeamstopDirection,
        setBeamstopDirection,
        delBeamstopDirection,
        "Property for beamstopDirection",
    )
    # Methods and properties for the 'beamstopSize' attribute

    def getBeamstopSize(self):
        return self._beamstopSize

    def setBeamstopSize(self, beamstopSize):
        if beamstopSize is None:
            self._beamstopSize = None
        elif beamstopSize.__class__.__name__ == "XSDataDouble":
            self._beamstopSize = beamstopSize
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor.setBeamstopSize argument is not XSDataDouble but %s"
                % beamstopSize.__class__.__name__
            )
            raise Exception(strMessage)

    def delBeamstopSize(self):
        self._beamstopSize = None

    beamstopSize = property(
        getBeamstopSize, setBeamstopSize, delBeamstopSize, "Property for beamstopSize"
    )
    # Methods and properties for the 'beamstopDistance' attribute

    def getBeamstopDistance(self):
        return self._beamstopDistance

    def setBeamstopDistance(self, beamstopDistance):
        if beamstopDistance is None:
            self._beamstopDistance = None
        elif beamstopDistance.__class__.__name__ == "XSDataDouble":
            self._beamstopDistance = beamstopDistance
        else:
            strMessage = (
                "ERROR! XSDataInputControlDozor.setBeamstopDistance argument is not XSDataDouble but %s"
                % beamstopDistance.__class__.__name__
            )
            raise Exception(strMessage)

    def delBeamstopDistance(self):
        self._beamstopDistance = None

    beamstopDistance = property(
        getBeamstopDistance,
        setBeamstopDistance,
        delBeamstopDistance,
        "Property for beamstopDistance",
    )

    def export(self, outfile, level, name_="XSDataInputControlDozor"):
        showIndent(outfile, level)
        outfile.write(unicode("<%s>\n" % name_))
        self.exportChildren(outfile, level + 1, name_)
        showIndent(outfile, level)
        outfile.write(unicode("</%s>\n" % name_))

    def exportChildren(self, outfile, level, name_="XSDataInputControlDozor"):
        XSDataInput.exportChildren(self, outfile, level, name_)
        if self._template is not None:
            self.template.export(outfile, level, name_="template")
        else:
            warnEmptyAttribute("template", "XSDataString")
        if self._first_image_number is not None:
            self.first_image_number.export(outfile, level, name_="first_image_number")
        else:
            warnEmptyAttribute("first_image_number", "XSDataInteger")
        if self._last_image_number is not None:
            self.last_image_number.export(outfile, level, name_="last_image_number")
        else:
            warnEmptyAttribute("last_image_number", "XSDataInteger")
        if self._first_run_number is not None:
            self.first_run_number.export(outfile, level, name_="first_run_number")
        else:
            warnEmptyAttribute("first_run_number", "XSDataInteger")
        if self._last_run_number is not None:
            self.last_run_number.export(outfile, level, name_="last_run_number")
        else:
            warnEmptyAttribute("last_run_number", "XSDataInteger")
        if self._line_number_of is not None:
            self.line_number_of.export(outfile, level, name_="line_number_of")
        else:
            warnEmptyAttribute("line_number_of", "XSDataInteger")
        if self._reversing_rotation is not None:
            self.reversing_rotation.export(outfile, level, name_="reversing_rotation")
        else:
            warnEmptyAttribute("reversing_rotation", "XSDataBoolean")
        if self._pixelMin is not None:
            self.pixelMin.export(outfile, level, name_="pixelMin")
        if self._pixelMax is not None:
            self.pixelMax.export(outfile, level, name_="pixelMax")
        if self._beamstopDirection is not None:
            self.beamstopDirection.export(outfile, level, name_="beamstopDirection")
        if self._beamstopSize is not None:
            self.beamstopSize.export(outfile, level, name_="beamstopSize")
        if self._beamstopDistance is not None:
            self.beamstopDistance.export(outfile, level, name_="beamstopDistance")

    def build(self, node_):
        for child_ in node_.childNodes:
            nodeName_ = child_.nodeName.split(":")[-1]
            self.buildChildren(child_, nodeName_)

    def buildChildren(self, child_, nodeName_):
        if child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "template":
            obj_ = XSDataString()
            obj_.build(child_)
            self.setTemplate(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "first_image_number":
            obj_ = XSDataInteger()
            obj_.build(child_)
            self.setFirst_image_number(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "last_image_number":
            obj_ = XSDataInteger()
            obj_.build(child_)
            self.setLast_image_number(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "first_run_number":
            obj_ = XSDataInteger()
            obj_.build(child_)
            self.setFirst_run_number(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "last_run_number":
            obj_ = XSDataInteger()
            obj_.build(child_)
            self.setLast_run_number(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "line_number_of":
            obj_ = XSDataInteger()
            obj_.build(child_)
            self.setLine_number_of(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "reversing_rotation":
            obj_ = XSDataBoolean()
            obj_.build(child_)
            self.setReversing_rotation(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "pixelMin":
            obj_ = XSDataInteger()
            obj_.build(child_)
            self.setPixelMin(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "pixelMax":
            obj_ = XSDataInteger()
            obj_.build(child_)
            self.setPixelMax(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "beamstopDirection":
            obj_ = XSDataString()
            obj_.build(child_)
            self.setBeamstopDirection(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "beamstopSize":
            obj_ = XSDataDouble()
            obj_.build(child_)
            self.setBeamstopSize(obj_)
        elif child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "beamstopDistance":
            obj_ = XSDataDouble()
            obj_.build(child_)
            self.setBeamstopDistance(obj_)
        XSDataInput.buildChildren(self, child_, nodeName_)

    # Method for marshalling an object
    def marshal(self):
        oStreamString = StringIO()
        oStreamString.write(unicode('<?xml version="1.0" ?>\n'))
        self.export(oStreamString, 0, name_="XSDataInputControlDozor")
        oStringXML = oStreamString.getValue()
        oStreamString.close()
        return oStringXML

    # Only to export the entire XML tree to a file stream on disk
    def exportToFile(self, _outfileName):
        outfile = open(_outfileName, "w")
        outfile.write(unicode('<?xml version="1.0" ?>\n'))
        self.export(outfile, 0, name_="XSDataInputControlDozor")
        outfile.close()

    # Deprecated method, replaced by exportToFile
    def outputFile(self, _outfileName):
        print(
            "WARNING: Method outputFile in class XSDataInputControlDozor is deprecated, please use instead exportToFile!"
        )
        self.exportToFile(_outfileName)

    # Method for making a copy in a new instance
    def copy(self):
        return XSDataInputControlDozor.parseString(self.marshal())

    # Static method for parsing a string
    def parseString(_inString):
        doc = minidom.parseString(_inString)
        rootNode = doc.documentElement
        rootObj = XSDataInputControlDozor()
        rootObj.build(rootNode)
        # Check that all minOccurs are obeyed by marshalling the created object
        oStreamString = StringIO()
        rootObj.export(oStreamString, 0, name_="XSDataInputControlDozor")
        oStreamString.close()
        return rootObj

    parseString = staticmethod(parseString)
    # Static method for parsing a file

    def parseFile(_inFilePath):
        doc = minidom.parse(_inFilePath)
        rootNode = doc.documentElement
        rootObj = XSDataInputControlDozor()
        rootObj.build(rootNode)
        return rootObj

    parseFile = staticmethod(parseFile)


# end class XSDataInputControlDozor


class XSDataResultControlDozor(XSDataResult):
    def __init__(self, status=None, imageDozor=None):
        XSDataResult.__init__(self, status)
        if imageDozor is None:
            self._imageDozor = []
        elif imageDozor.__class__.__name__ == "list":
            self._imageDozor = imageDozor
        else:
            strMessage = (
                "ERROR! XSDataResultControlDozor constructor argument 'imageDozor' is not list but %s"
                % self._imageDozor.__class__.__name__
            )
            raise Exception(strMessage)

    # Methods and properties for the 'imageDozor' attribute
    def getImageDozor(self):
        return self._imageDozor

    def setImageDozor(self, imageDozor):
        if imageDozor is None:
            self._imageDozor = []
        elif imageDozor.__class__.__name__ == "list":
            self._imageDozor = imageDozor
        else:
            strMessage = (
                "ERROR! XSDataResultControlDozor.setImageDozor argument is not list but %s"
                % imageDozor.__class__.__name__
            )
            raise Exception(strMessage)

    def delImageDozor(self):
        self._imageDozor = None

    imageDozor = property(
        getImageDozor, setImageDozor, delImageDozor, "Property for imageDozor"
    )

    def addImageDozor(self, value):
        if value is None:
            strMessage = (
                "ERROR! XSDataResultControlDozor.addImageDozor argument is None"
            )
            raise Exception(strMessage)
        elif value.__class__.__name__ == "XSDataControlImageDozor":
            self._imageDozor.append(value)
        else:
            strMessage = (
                "ERROR! XSDataResultControlDozor.addImageDozor argument is not XSDataControlImageDozor but %s"
                % value.__class__.__name__
            )
            raise Exception(strMessage)

    def insertImageDozor(self, index, value):
        if index is None:
            strMessage = "ERROR! XSDataResultControlDozor.insertImageDozor argument 'index' is None"
            raise Exception(strMessage)
        if value is None:
            strMessage = "ERROR! XSDataResultControlDozor.insertImageDozor argument 'value' is None"
            raise Exception(strMessage)
        elif value.__class__.__name__ == "XSDataControlImageDozor":
            self._imageDozor[index] = value
        else:
            strMessage = (
                "ERROR! XSDataResultControlDozor.addImageDozor argument is not XSDataControlImageDozor but %s"
                % value.__class__.__name__
            )
            raise Exception(strMessage)

    def export(self, outfile, level, name_="XSDataResultControlDozor"):
        showIndent(outfile, level)
        outfile.write(unicode("<%s>\n" % name_))
        self.exportChildren(outfile, level + 1, name_)
        showIndent(outfile, level)
        outfile.write(unicode("</%s>\n" % name_))

    def exportChildren(self, outfile, level, name_="XSDataResultControlDozor"):
        XSDataResult.exportChildren(self, outfile, level, name_)
        for imageDozor_ in self.getImageDozor():
            imageDozor_.export(outfile, level, name_="imageDozor")

    def build(self, node_):
        for child_ in node_.childNodes:
            nodeName_ = child_.nodeName.split(":")[-1]
            self.buildChildren(child_, nodeName_)

    def buildChildren(self, child_, nodeName_):
        if child_.nodeType == Node.ELEMENT_NODE and nodeName_ == "imageDozor":
            obj_ = XSDataControlImageDozor()
            obj_.build(child_)
            self.imageDozor.append(obj_)
        XSDataResult.buildChildren(self, child_, nodeName_)

    # Method for marshalling an object
    def marshal(self):
        oStreamString = StringIO()
        oStreamString.write(unicode('<?xml version="1.0" ?>\n'))
        self.export(oStreamString, 0, name_="XSDataResultControlDozor")
        oStringXML = oStreamString.getValue()
        oStreamString.close()
        return oStringXML

    # Only to export the entire XML tree to a file stream on disk
    def exportToFile(self, _outfileName):
        outfile = open(_outfileName, "w")
        outfile.write(unicode('<?xml version="1.0" ?>\n'))
        self.export(outfile, 0, name_="XSDataResultControlDozor")
        outfile.close()

    # Deprecated method, replaced by exportToFile
    def outputFile(self, _outfileName):
        print(
            "WARNING: Method outputFile in class XSDataResultControlDozor is deprecated, please use instead exportToFile!"
        )
        self.exportToFile(_outfileName)

    # Method for making a copy in a new instance
    def copy(self):
        return XSDataResultControlDozor.parseString(self.marshal())

    # Static method for parsing a string
    def parseString(_inString):
        doc = minidom.parseString(_inString)
        rootNode = doc.documentElement
        rootObj = XSDataResultControlDozor()
        rootObj.build(rootNode)
        # Check that all minOccurs are obeyed by marshalling the created object
        oStreamString = StringIO()
        rootObj.export(oStreamString, 0, name_="XSDataResultControlDozor")
        oStreamString.close()
        return rootObj

    parseString = staticmethod(parseString)
    # Static method for parsing a file

    def parseFile(_inFilePath):
        doc = minidom.parse(_inFilePath)
        rootNode = doc.documentElement
        rootObj = XSDataResultControlDozor()
        rootObj.build(rootNode)
        return rootObj

    parseFile = staticmethod(parseFile)


# end class XSDataResultControlDozor


# End of data representation classes.



import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject


class RobodiffMotor(HardwareObject):
    (NOTINITIALIZED, UNUSABLE, READY, MOVESTARTED, MOVING, ONLIMIT) = (0, 1, 2, 3, 4, 5)

    def __init__(self, name):
        super().__init__(name)
        self.__initialized = False

    def init(self):
        self.motorState = RobodiffMotor.NOTINITIALIZED
        self.username = self.actuator_name
        # gevent.spawn_later(1, self.end_init)

    def end_init(self):
        if self.__initialized:
            return
        controller = HWR.get_hardware_repository().get_hardware_object(
            self.get_property("controller")
        )  # self.get_object_by_role("controller")

        # this is ugly : I added it to make the centring procedure happy
        self.specName = self.actuator_name

        self.motor = getattr(controller, self.actuator_name)
        self.connect(self.motor, "position", self.positionChanged)
        self.connect(self.motor, "state", self.updateState)
        self.__initialized = True

    def connect_notify(self, signal):
        if signal == "valueChanged":
            self.emit("valueChanged", (self.get_value(),))
        elif signal == "stateChanged":
            self.updateState(emit=True)
        elif signal == "limitsChanged":
            self.motorLimitsChanged()

    def updateState(self, state=None, emit=False):
        self.end_init()
        if state is None:
            state = self.motor.state()
        # convert from grob state to Hardware Object motor state
        if state == "MOVING":
            state = RobodiffMotor.MOVING
        elif state == "READY":
            state = RobodiffMotor.READY
        elif state == "ONLIMIT":
            state = RobodiffMotor.ONLIMIT
        else:
            state = RobodiffMotor.UNUSABLE

        self.set_is_ready(state > RobodiffMotor.UNUSABLE)

        if self.motorState != state:
            self.motorState = state
            emit = True
        if emit:
            self.emit("stateChanged", (self.motorState,))

    def get_state(self):
        self.end_init()
        self.updateState()
        return self.motorState

    def motorLimitsChanged(self):
        self.end_init()
        self.emit("limitsChanged", (self.get_limits(),))

    def get_limits(self):
        self.end_init()
        return self.motor.limits()

    def positionChanged(self, absolutePosition):
        # print self.name(), absolutePosition
        self.emit("valueChanged", (absolutePosition,))

    def get_value(self):
        self.end_init()
        return self.motor.position()

    def _set_value(self, value):
        self.end_init()
        self.updateState("MOVING")
        self.motor.move(value, wait=False)

    def waitEndOfMove(self, timeout=None):
        with gevent.Timeout(timeout):
            self.motor.wait_move()

    def motorIsMoving(self):
        self.end_init()
        return self.motorState == RobodiffMotor.MOVING

    def get_motor_mnemonic(self):
        self.end_init()
        return self.actuator_name

    def stop(self):
        self.end_init()
        self.motor.stop()



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Abstract Characterisation class.
Defines:
characterise, get_html_report methods, input_from_params, dc_from_output and
get_default_characterisation_parameters abstract methods;
prepare_input and is_running methods.
"""

import abc

import gevent.event

from mxcubecore.BaseHardwareObjects import HardwareObject


class AbstractCharacterisation(HardwareObject):
    """Abstract Characterisation Class"""

    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        super().__init__(name)
        self.processing_done_event = gevent.event.Event()

    @abc.abstractmethod
    def characterise(self, _input):
        """Start the characterosation.
        Args:
            input (object) Characterisation input object.
        Returns:
            (str) The Characterisation result.
        """

    @abc.abstractmethod
    def get_html_report(self, output):
        """Get the path to the characterisation html report, generated by the
        characterisation softare.
        Args:
            output (object): Characterisation output object
        Returns:
            (str) The path to the html result report.
        """

    @abc.abstractmethod
    def input_from_params(self, ref_parameters, char_params, path_str):
        """
        Args:
            ref_parameters: (A named tuple or object with following fields):
              'id',
              'prefix',
              'run_number',
              'template',
              'first_image',
              'num_images',
              'osc_start',
              'osc_range',
              'overlap',
              'exp_time',
              'num_passes',
              'comments',
              'path',
              'centred_positions',
              'energy',
              'resolution',
              'transmission',
              'shutterless',
              'inverse_beam',
              'screening_id'

        Args:
            char_params: (A named tuple or object with following fields):
              # Optimisation parameters
              'aimed_resolution'
              'aimed_multiplicity'
              'aimed_i_sigma'
              'aimed_completness'
              'strategy_complexity'
              'induce_burn'
              'use_permitted_rotation'
              'permitted_phi_start'
              'permitted_phi_end'

              # Crystal
              'max_crystal_vdim'
              'min_crystal_vdim'
              'max_crystal_vphi'
              'min_crystal_vphi'
              'space_group'

              # Characterisation type
              'use_min_dose'
              'use_min_time'
              'min_dose'
              'min_time'
              'account_rad_damage'
              'not_use_low_res'
              'auto_res'
              'opt_sad'
              'sad_res'
              'determine_rad_params'

              # Radiation damage model
              'rad_suscept'
              'beta'
              'sigma'

        Args:
          path_str (Template string representing path to each image)

        Returns:
            Input for characterisation software
        """

    @abc.abstractmethod
    def dc_from_output(self, output):
        """Create a data collection from characterisation result
        Args:
            output (object) Characterisation result object
        Returns:
            (queue_model_objects.DataCollection)
        """

    @abc.abstractmethod
    def get_default_characterisation_parameters(self, default_input_file):
        """
        Args:
            default_input_file (str): Path to file containing default input.
        Returns:
            (queue_model_objects.CharacterisationsParameters): object with
                 default parameters.
        """

    def prepare_input(self, _input):
        """Called by characterise before characterisation starts, can be used
        to manipulate edna_input object before characterisation starts.
        Example: to set a site specific output directory
        Args:
           _input (object) Characterisation input object
        """

    def is_running(self):
        """
        Returns:
            (bool) True if process is running otherwise False
        """
        return not self.processing_done_event.is_set()
